/* dbping2.ec - Improved version!
 */
/*
  RCS Header:
  -----------
  $Header: /home/art/GoogleDrive/Latest\040Art's\040Stuff/RCS/dbping.ec,v 1.17 2023/12/21 16:26:48 art Exp $

  RCS Log:
  --------
  $Log: dbping.ec,v $
  Revision 1.17  2023/12/21 16:26:48  art
  Made the arguments more like other utilities while still supporting 
  the original command line.

  Revision 1.16  2021/12/21 17:12:40  art
  Updated Copyright notice.

  Revision 1.15  2021/06/09 19:49:23  art
  Updated so that the -l/-c combination executes the specified command
  in the background if the limit timer expires before the connection
  completes allowing for collecting data about the delay such as
  from ifxcollect or simple onstats.

  Revision 1.14  2021/06/02 13:49:25  art
  Added new -l limit and -c command options to execute a command
  if the connection time exceeds the specified limit. Passing -l
  will increase the timeout to the limit plus 1 second if the
  timeout is shorter than the limit.

  Revision 1.13  2020/11/19 21:49:32  art
  Made inlude changes for better Solaris support.

  Revision 1.12  2020/05/15 16:10:49  art
  Improved usage.

  Revision 1.11  2020/05/15 16:01:07  art
  Added -u user, -p password, & -t timeout options.

  Revision 1.10  2019/03/15 15:17:23  art
  Added missing header file to resolve new GCC warning.

  Revision 1.9  2017/12/26 16:54:33  art
  Fixed a couple of printf's

  Revision 1.8  2016/02/05 15:42:26  art
  Fixed typo.

  Revision 1.7  2016/02/05 15:26:58  art
  Added -V and copyright notice. Added a call to ifx_getcur_conn_name() to get the alias
  that was actually connected.

  Revision 1.6  2014/09/14 16:44:15  art
  Fixed problems with calling localtime inside asctime on some systems.

  Revision 1.5  2001/12/10 21:11:14  kagel
  Removed outdated comments.

  Revision 1.4  2001/12/10 20:16:16  kagel
  Added auto connect to database sysmaster if none provided.

  Revision 1.2  2001/12/10 20:11:58  kagel
  Improved.

  Revision 1.2  2000/11/16 17:30:02  kagel
  Updated to handle refused connections.

  Revision 1.1  2000/11/16 17:15:42  kagel
  Initial revision


*/

static const char RCSHeader[] = "@(#)$Header: /home/art/GoogleDrive/Latest\040Art's\040Stuff/RCS/dbping.ec,v 1.17 2023/12/21 16:26:48 art Exp $";
static const char RCSWhat[] = "$What: <@(#) dbping2.ec,v	 1.5 > $ $Date: 2023/12/21 16:26:48 $";
static const char RCSCompile[] = "@(#)$cc: " __FILE__ " " __DATE__ " " __TIME__ " $";

/* dbping.ec - try to connect to the default or specified server and report
   timing.
*/

#include <stdio.h>
#include <fcntl.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#if defined( __svr4__ )
#include <signal.h>
#else
#include <sys/signal.h>
#endif
#include <time.h>
#include <sys/time.h>
#include <getopt.h>
#include <errno.h>

#define SQLOK 0  
char command[4096] = "onstat -a";
int timeout = 5;

void handler( int signal_num ); 
void run_cmd( int status );

EXEC SQL include sqlca ;
EXEC SQL include datetime ;
EXEC SQL include locator ; 
EXEC SQL include sqltypes ; 

struct timeval tp, tp1, tp2, tp3 ; 

void usage( char *prog )
{
    fprintf( stderr, "Usage: %s [-u user -p pwd] [-t timeout] [[database/server spec]\n\n", prog );
    fprintf( stderr, "\t   -V - Print version and copyright info.\n" );
    fprintf( stderr, "\t   -H - Server connection name.\n" );
    fprintf( stderr, "\t   -d - Database name.\n" );
    fprintf( stderr, "\t   -u - Supply user for non-trusted connections. (Requires -p)\n" );
    fprintf( stderr, "\t   -p - Supply password for non-trusted connections. (Requires -u)\n" );
    fprintf( stderr, "\t   -t - Supply timeout. (Default: 5 seconds)\n" );
    fprintf( stderr, 
	     "\t   -l - Limit in fractional seconds. Pings exceeding this will execute\n"
	     "\t        the command.\n" );
    fprintf( stderr,
	     "\t   -c - Command to execute if the time limit is exceeded. Must be a single\n"
	     "\t        word. Specify a script if you need command line arguments passed.\n" );
    fprintf( stderr, "\t   -h - Print this usage.\n" );
    fprintf( stderr,
	     "\nProvide a server connection name via the INFORMIX_SERVER environment variable,\n"
	     "the -H option, or as part of the database name (ex: mydb@myserver). Database\n"
	     "can be provided as the last argument on the command line or using the -d flag\n" );
}

int main(int argc, char *argv[]) {

    EXEC SQL BEGIN DECLARE SECTION ; 
    string database[1025], server[1025], connected_name[1025], hostname[1025], connection[1025];
    string user[33], passwd[128];
    EXEC SQL END DECLARE SECTION ; 
    struct tm *temptm;
    int v, ercd, ch;
    double limit = 99999.99;
    double pingt;
    int have_user = 0, have_pwd = 0, have_server = 0, have_database = 0;
    int have_timeout = 0, have_limit = 0, have_command = 0, have_connection = 0;

    connection[0] = (char)0;
    
    while ((ch = getopt(argc, argv, "Vp:u:t:l:d:c:hH:?")) != EOF){
        switch(ch) {
	case 'H':
	    strncpy( connection, optarg, sizeof connection );
	    have_connection = 1;
	    break;
	case 'd':
	    strncpy( database, optarg, sizeof database );
	    if (strchr( database, '@' )) {
		strcpy( server, database );
		have_server = 1;
	    }
	    have_database = 1;
	    break;
	case 'p':
	    strncpy( passwd, optarg, sizeof passwd );
	    have_pwd = 1;
	    break;
	case 'u':
	    strncpy( user, optarg, sizeof user );
	    have_user = 1;
	    break;
	case 'l':
	    limit = atof( optarg );
	    have_limit = 1;
	    fprintf( stderr, "Run limit set to %f seconds.\n", limit );
	    break;
	case 'c':
	    strncpy( command, optarg, sizeof command );
	    have_command = 1;
	    break;
	case 't':
	    timeout = atol( optarg );
	    have_timeout = 1;
	    break;
	case 'V':
	    fprintf( stderr,
		     "Version: %s.  Copyright 1990-2024 Art S. Kagel.\n",
		     "$Revision: 1.17 $" );
	    return 0;
	default:
	    usage( argv[0] );
	    return 0;
	}
    }

    if (have_limit && (limit > (double)timeout))
	timeout = (int)(limit + 1.0);

    if ((have_user + have_pwd) == 1) {
	fprintf( stderr, "If -p or -u are provided, both are required!\n\n" );
	usage( argv[0] );
	return 1;
    }

    if (argc > optind) {
	char *loc;

	if (strlen( argv[optind] ) > 1024) {
	    fprintf( stderr, "Invalid server: %s.\n", argv[1] );
	    usage( argv[0] );
	    exit(1);
	}
	if (atol( argv[optind] ) == 0) {
	    if ((loc = strchr( argv[optind], '@' ))) {
		if (loc == argv[optind]) {
		    strcpy( connection, (argv[optind]+1) );
		    have_connection = 1;
		} else {
		    strcpy( server, argv[optind] );
		    have_server = 1;
		}
	    } else {
		strcpy( database, argv[optind] );
		have_database = 1;
	    }
	    optind++;
	}
    }
    if (have_server == 0 && have_connection == 0) {
	if (sprintf( connection, "%s", getenv( "INFORMIXSERVER" ) ) > 0)
	    have_connection = 1;
	else {
	    fprintf( stderr, "No server connection provided!\n" );
	    exit(1);
	}
    }

    if (have_server == 0 && have_database == 0) {
	strcpy( database, "sysmaster" );
    }

    if (have_server == 0)
	snprintf( server, sizeof server, "%s@%s", database, connection );

    if (argc > optind && timeout == 0) {
	timeout = atol( argv[optind] );
    }

    if (!timeout)
	timeout = 5;

    gettimeofday( &tp, (struct timezone *)NULL );
    temptm = localtime( (time_t *)&(tp.tv_sec) );
    fprintf( stderr,
	     "%s started with %d second timeout at %s\n",
	     argv[0],
	     timeout,
	     asctime( temptm ) );

    if (have_limit) {
	struct itimerval curr;

	/* Set up a one-shot micro timeout to launch the command. */
	curr.it_value.tv_sec = (int)limit;
	curr.it_value.tv_usec = (int)((limit -(int)limit) * 1000000);
	curr.it_interval.tv_sec = 0;
	curr.it_interval.tv_usec = 0;
	if ((ercd = setitimer( ITIMER_REAL, &curr, (struct itimerval *)NULL )) != 0) {
	    fprintf( stderr, "setitimer failed with: %d. Errno = %d.\n", ercd, errno );
	} else {
	    fprintf( stderr,
		     "setitimer succeeded. Set to %ld seconds & %ld microseconds.\n",
		     curr.it_value.tv_sec,
		     curr.it_value.tv_usec );
	}

	signal( SIGALRM, run_cmd );
    }

    if (have_user == 1) {
	printf( " EXEC SQL CONNECT TO %s USER %s USING %s;\n", server, user, passwd );
	EXEC SQL CONNECT TO :server USER :user USING :passwd;
    } else {
	printf( " EXEC SQL CONNECT TO %s;\n", server );
	EXEC SQL CONNECT TO :server;
    }
    
    gettimeofday( &tp2, (struct timezone *)NULL );

    tp3.tv_sec = tp2.tv_sec - tp.tv_sec;
    tp3.tv_usec = tp2.tv_usec - tp.tv_usec;
    if (tp3.tv_usec < 0) {
	tp3.tv_sec--;
	tp3.tv_usec += 1000000;
    }
    pingt = ((double)(tp3.tv_sec) + (((double)tp3.tv_usec)/1000000.00));

    if (sqlca.sqlcode < 0) {
        temptm = localtime( (time_t *)&(tp.tv_sec) );
        fprintf( stderr,
	         "Connection failed in %d.%6.6d seconds, at: %s",
	         (int)tp3.tv_sec,
	         (int)tp3.tv_usec,
	         asctime( temptm ) );
	fprintf( stderr,
		 "Engine is available but connection failed: %d, %d.\n",
		 sqlca.sqlcode,
		 sqlca.sqlerrd[1] );
	return 2;
    }
    temptm = localtime( (time_t *)&(tp.tv_sec) );
    fprintf( stderr,
	     "Connection successful in %d.%6.6d seconds, at: %s",
	     (int)tp3.tv_sec,
	     (int)tp3.tv_usec,
	     asctime( temptm ) );
    /* Moved command run to an async child process so it executes during the delay. */
    /* if (have_limit && pingt > limit) { */
    /* 	fprintf( stderr, "Time limit of %f second exceeded! Executing command.\n", limit ); */
    /* 	system( command ); */
    /* } */

 again:
    EXEC SQL SELECT cf_effective, dbinfo( 'dbhostname' )
	INTO :connected_name, :hostname
	FROM sysmaster:sysconfig WHERE cf_name = 'DBSERVERNAME';
 if (sqlca.sqlcode == -349) {
     EXEC SQL DATABASE sysmaster;
     goto again;
 } else if (sqlca.sqlcode != 0) {
     fprintf( stderr,
	      "Error getting servername: %d, ISAM=%d.\n",
	      sqlca.sqlcode,
	      sqlca.sqlerrd[1] );
     return 1;
 } else {
     fprintf( stderr,
	      "Actually connected to: %s on %s.\n",
	      connected_name,
	      hostname );
     fprintf( stderr,
	      "\tConnection named: %s.\n",
	      ifx_getcur_conn_name() );
 }

 return 0;
}

void handler( int signal_num ) { 
    struct tm *temptm;

    if (signal_num != SIGALRM) return;
    
    gettimeofday( &tp2, (struct timezone *)NULL );
    tp3.tv_sec = tp2.tv_sec - tp.tv_sec;
    tp3.tv_usec = tp2.tv_usec - tp.tv_usec;
    if (tp3.tv_usec < 0) {
	tp3.tv_sec--;
	tp3.tv_usec += 1000000;
    }
    temptm = localtime( (time_t *)&(tp.tv_sec) );
    fprintf( stderr,
	     "Connection timed out after %d.%6.6d seconds, at: %s",
	     (int)tp3.tv_sec,
	     (int)tp3.tv_usec,
	     asctime( temptm ) );

    exit(1);
}

void run_cmd( int sig )
{
    unsigned long pid, ppid, cpid;
    char shell[1025];
    
    if (sig != SIGALRM) return;
    
    ppid = getpid();
    pid = fork();
    cpid = getpid();
    if (ppid != cpid) {
	// In the child. Execute the cmd here.
	fprintf( stderr, "From child(%lu);\n", cpid );
	strncpy( shell, getenv( "SHELL" ), 1025 );
	execlp( command, command, (char *)NULL );
	if (errno != 0) {
	    fprintf( stderr, "exec(%s) failed with errno = %d.\n", command, errno );
	    return;
	}
    } else {
	fprintf( stderr, "From parent (%lu).\n", ppid );
	/* set a signal handler for the timeout timer */ 
	signal( SIGALRM, handler ) ; 
	alarm( timeout );
    }
    return;
}
