/*
  Program Id: dbstruct

  Program Description: this program will generate a "C" struct for a
  specified database or a set of tables. 

  Author: Art S. Kagel from schema.ec 
  originally taken from Rick Friedman's dbschema_cc for Sybase

  Date: May of 1992  

  Version: 1.0

  Comments:    

  Revisions:

  Usage:    dbstruct [-F] [-h hostname] -d databasename [-t tablename] [filename]

  RCS Header
  ----------
  $Header: /home/art/GoogleDrive/Latest\040Art's\040Stuff/RCS/dbstruct.ec,v 2.2 2022/03/15 15:17:09 art Exp $

  RCS Log
  -------
  $Log: dbstruct.ec,v $
  Revision 2.2  2022/03/15 15:17:09  art
  Updated the usage.

  Revision 2.1  2022/03/15 14:57:32  art
  Updated copyright notice.

  Revision 2.0  2022/03/15 14:55:16  art
  Recoded using switches and enum to make adding new output options 
  easier. Added new -R option to generate ROW TYPE definition.

  Revision 1.30  2021/12/21 17:09:52  art
  Added -l option to display a comma separated column list
  suitable for projection clauses etc.

  Revision 1.30  2021/04/27 16:44:22  art
  Improved messages for unsupported data types (mostly UDTs).
  Added new -l mode to generate a simple comma separated 
  column name list suitable for projection clauses or code.

  Revision 1.29  2019/03/15 15:07:15  art
  Fixed some compiler warnings and updated inlines to work with newer GCC compiler versions.

  Revision 1.28  2016/09/06 18:34:15  art
  Updated usage() function.

  Revision 1.27  2016/09/06 17:34:33  art
  Fixed extended types in FORTRAN and 4GL - as well as can be done.
  Fixed a couple of minor issues with the new features.

  Revision 1.26  2016/09/01 17:15:02  art
  Fixed UDT and extended types output.
  Added new format -D to generate simple SQL DDL.
  Added new option -M to generate each table's output to a separate file
  named after the database.table.ext. If -M is passed the optional argument
  overrides the default extension which is appropriate to the structure
  type.

  Revision 1.25  2016/06/27 21:18:47  art
  Updated copyright notice.

  Revision 1.24  2015/10/12 12:20:27  art
  Typos.

  Revision 1.23  2015/10/12 12:17:23  art
  Added support for BOOLEAN type.

  Revision 1.22  2014/05/15 21:21:42  art
  Added options to ignore shadow columns like ERKEY, VERCOLS, & REPLCHECK
  columns. -k, -v, & -r respectively

  Revision 1.21  2013/12/24 17:12:57  art
   Fixed compile time warnings on HPUX.  Updated copyright notice.

  Revision 1.20  2013/07/09 16:17:06  art
  Fixed problem with old 18byte object names.  Added support for 128byte names.
  Added support for INT8/SERIAL8 & BOOLEAN types

  Revision 1.19  2013/02/10 22:18:19  art
  Updated Copyright

  Revision 1.18  2012/05/30 01:25:33  art
  Protected TRUE and FALSE from pre-definition (mostly on AIX).

  Revision 1.17  2011/06/20 16:45:36  art
  Fixed output broken when 4GL support was added.

  Revision 1.16  2011/05/23 16:04:57  art
  Updated the copyright statement.

  Revision 1.15  2011/04/12 18:46:34  art
  Added support for 4GL records.

  Revision 1.14  2009/11/18 20:04:35  art
  Added support for external tables and synonyms.

  Revision 1.13  2009/07/23 17:29:24  art
  Missing type for static.  Some compilers assume 'int' others complain. Fixed.

  Revision 1.12  2009/07/19 14:49:52  art
  Added protections on use of BIGINT/BIGSERIAL for older esql/C versions.

  Revision 1.11  2009/06/24 02:01:21  art
  Fixed mixed case code.

  Revision 1.10  2009/05/13 23:53:11  art
  *** empty log message ***

  Revision 1.9  2009/05/13 23:51:53  art
  Made to work for mixed case table name arguments when DELIMIDENT is disabled.

  Revision 1.8  2009/04/12 21:58:02  art
  Updated usage.

  Revision 1.7  2009/04/12 21:56:12  art
  Removed BB-isms.  Added Revision and copyright printout.

  Revision 1.6  2009/04/01 16:57:23  art
  Added support for bigint, bigserial, and lvarchar types.

  Revision 1.5  2004/01/14 16:54:48  kagel
  Fixed return type and prototype of rexit.

  Revision 1.4  2003/12/01 21:56:59  kagel
  Fixed portability to IBM.
  AIX.

  Revision 1.3  1996/08/06 18:39:52  kagel
  Added -B flag to represent DATETIME and INTERVAL as Bloomberg Datetime.

  # Revision 1.2  1995/05/04  15:07:53  kagel
  # Added ESQL and FORTRAN support I think?!?!?!
  #
  # Revision 1.1  1994/09/02  15:16:09  kagel
  # Initial revision
  #
  # Revision 1.6  1993/07/29  19:51:52  kagel
  # Added RCS headers.

  Revision 1.2  93/01/17  09:52:50  ask
  Missed a change to the extent size warning.

  Revision 1.1  93/01/17  09:47:26  ask
  Initial revision


*/
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <sys/types.h>
#if defined(__sysv__) || defined(i386) || defined(_AIX) || defined(__GNUC__)
#include <string.h>
#else
#include <strings.h>
#endif
#include <libgen.h>
#include "getopt.h"

EXEC SQL INCLUDE sqlca;
EXEC SQL INCLUDE sqltypes;
EXEC SQL INCLUDE datetime;

static char rcs_header[] = "$Header: /home/art/GoogleDrive/Latest\040Art's\040Stuff/RCS/dbstruct.ec,v 2.2 2022/03/15 15:17:09 art Exp $";

extern struct sqlca_s sqlca;

EXEC SQL BEGIN DECLARE SECTION;

struct { 
    char tabname[129];
    long tabid;
    long partnum, fextsize, nextsize;
    char locklevel[2];
    long npused;
    char tabtype[2];
    int nrows;
} systables, prev;

struct {
    long tabid;
    char colname[129];
    short colno;
    short coltype;
    short collength;
    int colmin;
    int colmax;
} syscolumns;

string databasename[129], host[100], statement[2048], connection[256];
string tablename[129], xtdname[129], xtddesc[257], dbspace[257];
short xtdind;
int xtdtype;

EXEC SQL END DECLARE SECTION;

extern int optind;
extern char *optarg;

#if !defined(SQLINFXBIGINT)
#define SQLINFXBIGINT 52
#define SQLBIGSERIAL 53
#endif

#if !defined(SQLINT8)
#define SQLINT8 17
#define SQLSERIAL8 18
#endif

char *a_space;
int ptcnt[7] = { 4, 2, 2, 2, 2, 2, 0 };
char dtparts[7][10] = { "YEAR",
			"MONTH",
			"DAY",
			"HOUR",
			"MINUTE",
			"SECOND",
			"FRACTION" };
static char columns[1024][129];
int Have_FORTRAN = 0, Have_ESQL = 0, Have_FourGL = 0, Do_VCHAR = 1, Have_DDL = 0, Have_ROW = 0;

typedef enum { CLang, CPlus, FORTRAN, ESQL, FourGL, DDL, Listing, ROW } OutTypes ;
int OutType = CLang;

#define TBSFACT 16777216L
#if !defined( FALSE )
#define FALSE (0)
#endif
#if !defined( TRUE )
#define TRUE (!FALSE)
#endif

#if !defined( __GNUC__ )
#define strlower( t, s, m ) {					\
	char *target = (t), *source = (s); int max = (m);	\
	while (max-- && *source) {				\
	    *target++ = tolower(*source++);			\
	}							\
	*target = 0;						\
    }
#else
static inline void strlower( char *target, char *source, int max )
{
    while (max-- && *source) {
        *target++ = tolower(*source++);
    }
    *target = 0;
}
#endif

void rexit(int stat);
void usage( char *program );
void ucase( char *target, char *source );
char *typename( int coltype, int collength, char *xtdname, int xtdtype, char *xtddesc, char *dbspace );
char *typename_f( int coltype, int collength, char *xtdname, int xtdtype, char *xtddesc, char *dbspace );

int main(int argc, char *argv[])
{
    static int pass_no = 1;
    int i, j, c, errflag=0, dflag=0, tflag=0, hflag=0, allows_nulls;
    long colwidth, pt1, pt2, fr1, fr2;
    FILE *outfile = stdout;
    char filename[1025], upname[129], extension[101];
    static char tempstr[1024];
    int Delimident_Set = 0;
    int noReplCheck = 0, noVerCols = 0, noErKeys = 0;
    int multifile = 0;
    int do_listing = 0;

    prev.tabname[0] = (char)0;
    strcpy( tablename, "*" );
    host[0] = (char)0;

    if (getenv( "DELIMIDENT" ) != (char *)NULL) {
	Delimident_Set = 1;
    }

    while ((c = getopt(argc, argv, "d:t:h:DFGMPRVefklrv?")) != EOF){
        switch(c) {
	case 'V':
	    fprintf( stderr,
		     "dbstruct.ec Revision: %s.  Copyright 1990-2022, Art S. Kagel\n",
		     "$Revision: 2.2 $" );
	    return 0;
	case 'l':
	    OutType = Listing;
	    do_listing = 1;
	    break;
	case 'D':
	    OutType = DDL;
	    Have_DDL = 1;
	    break;
	case 'P':
	    OutType = CPlus;
	    break;
	case 'R':
	    OutType = ROW;
	    Have_ROW = 1;
	    break;
	case 'M':
	    multifile = 1;
	    break;
	case 'G':
	    OutType = FourGL;
	    Have_FourGL = 1;
	    break;
	case 'F':
	    OutType = FORTRAN;
	    Have_FORTRAN = 1;
	    break;
	case 'e':
	    OutType = ESQL;
	    Have_ESQL = 1;
	    break;
	case 'f':
	    Do_VCHAR = 0;
	    OutType = ESQL;
	    Have_ESQL = 1;
	    break;
	case 'd':
	    if (dflag) {
		errflag++;
	    } else {
		dflag++;
		strncpy(databasename, optarg, 128);
		databasename[128] = (char)0;
	    }
	    break;
	case 't':
	    if (tflag) {
		errflag++;
	    } else {
		tflag++;
		if (!Delimident_Set) {
		    strlower( tablename, optarg, 128 );
		} else {
		    strncpy(tablename, optarg, 128);
		    tablename[128] = (char)0;
		}
	    }
	    break;
	case 'h':
	    if (hflag) {
		errflag++;
	    } else {
		hflag++;
		strncpy(host, optarg, 100);
		host[99] = (char)0;
	    }
	    break;
	case 'v':
	    noVerCols = 1;
	    break;
	case 'r':
	    noReplCheck = 1;
	    break;
	case 'k':
	    noErKeys = 1;
	    break;
	case '?':
	default:
	    errflag++;
	    break;
        }
    }

    if ((Have_ROW + Have_FORTRAN + Have_FourGL + Have_DDL + Have_ESQL + do_listing) > 1) {
	fprintf( stderr, "Only one of -D, -F, -G, -R, -l, or -e/f permitted!\n" );
	usage( argv[0] );
        exit(1);
    }

    if (optind < argc) {
	if (multifile == 0) {
	    snprintf( filename, sizeof filename, argv[optind], 1024 );
	    outfile = fopen( filename, "w" );
	    if (!outfile) {
		fprintf( stderr, "Cannot open output: %s.\n", filename );
		exit(3);
	    }
	} else {
	    snprintf( extension, sizeof extension, "%s", argv[optind] );
	    outfile = (FILE *)NULL;
	}
    } else {
	switch (OutType) {
	case FORTRAN:
	    snprintf( extension, sizeof extension, "%s", "f77" );
	    break;
	case FourGL:
	    snprintf( extension, sizeof extension, "%s", "4gl" );
	    break;
	case ESQL:
	    snprintf( extension, sizeof extension, "%s", "ec" );
	    break;
	case DDL:
	case ROW:
	    snprintf( extension, sizeof extension, "%s", "sql" );
	    break;
	case Listing:
	    snprintf( extension, sizeof extension, "%s", "list" );
	    break;
	case CLang:
	    snprintf( extension, sizeof extension, "%s", "c" );
	    break;
	case CPlus:
	    snprintf( extension, sizeof extension, "%s", "c++" );
	    break;
	}
    }
    
    if (errflag || dflag == 0){
	usage( argv[0] );
        exit(1);
    }
    if(strcmp(databasename,"")==0 || ( tflag == 1 && strcmp(tablename,"")==0)){
	usage( argv[0] );
        exit(1);
    }

    strcpy( connection, databasename );
    if (hflag) {
	strcat( connection, "@" );
	strcat( connection, host );
	EXEC SQL
	    CONNECT TO :connection;
    } else {
	EXEC SQL DATABASE :connection;
    }
    if ( sqlca.sqlcode )
	{
	    fprintf( stderr,
		     "Cannot access host (%s) & database (%s); error=%d\n",
		     host,
		     databasename,
		     sqlca.sqlcode );
	    rexit(1);
	}

    /* Select for tables and columns. */
    strcpy( statement, "SELECT a.tabname, a.tabid, a.partnum, " );
    strcat( statement, "a.fextsize, a.nextsize" );
    strcat( statement, ", a.locklevel, b.colname, b.colno, b.coltype" );
    strcat( statement, ", b.collength, a.npused, a.tabtype " );
    strcat( statement,
	    ", colmin, colmax, nrows, c.name as xtdname, b.extended_id as xtdtype, d.description as xtddesc, " );
    strcat( statement,
	    "case WHEN (MOD(b.coltype, 256) IN (11, 12) AND b.colmin > 0) THEN "
	    "     (SELECT spacename FROM sysblobs sb WHERE sb.tabid = a.tabid and sb.colno = b.colno ) "
	    "     ELSE 'TABLE' "
	    "END AS dbspace " );
    strcat( statement, "FROM systables a "
			"LEFT OUTER JOIN syscolumns b "
			"  ON a.tabid = b.tabid "
			"LEFT OUTER JOIN sysxtdtypes c "
			"  ON b.extended_id = c.extended_id "
			"LEFT OUTER JOIN sysxtddesc d "
			"  ON c.extended_id = d.extended_id " );
    strcat( statement, "WHERE a.tabname MATCHES ? " );
    if (!tflag)
	strcat( statement, "AND a.tabid >= 100 " );
    if (noReplCheck)
	strcat ( statement, " AND b.colattr != 9 " );
    if (noErKeys)
	strcat ( statement, " AND (b.colattr != 17 AND b.colattr != 33 AND b.colattr != 65) " );
    if (noVerCols)
	strcat ( statement, " AND (b.colattr != 3 AND b.colattr != 5) " );
    strcat( statement, "ORDER BY a.tabid, b.colno; " );

    EXEC SQL
	PREPARE stmt1 FROM :statement;
    if (sqlca.sqlcode) {
	fprintf( stderr,
		 "Cannot PREPARE select statement for table data, error: %d.\n",
		 sqlca.sqlcode );
	fprintf( stderr, "Error at character %d.\nStatement is:\n",
		 sqlca.sqlerrd[4] );
	fprintf( stderr, "%s\n", statement );
	rexit(2);
    }

    EXEC SQL
	DECLARE a1 CURSOR FOR stmt1;
    if (sqlca.sqlcode ) {
        fprintf( stderr,
		 "Cannot declare a1 cursor; Error = %d\n", sqlca.sqlcode );
        rexit(1);
    }

    EXEC SQL
	OPEN a1 USING :tablename;
    if (sqlca.sqlcode == SQLNOTFOUND) {
	fprintf( stderr, "No table matching \"%s\".\n", tablename );
	rexit(3);
    } else if ( sqlca.sqlcode ) {
        fprintf( stderr,"Cannot open a1 cursor; Error = %d\n", sqlca.sqlcode );
        rexit(1);
    }

    i = 0;
    if (OutType == ESQL)
	fprintf( outfile, "EXEC SQL BEGIN DECLARE SECTION;\n" );
    for(;;) {
        EXEC SQL
	    fetch a1 into
            :systables.tabname,
	    :systables.tabid,
	    :systables.partnum,
	    :systables.fextsize,
	    :systables.nextsize,
	    :systables.locklevel,
            :syscolumns.colname,
            :syscolumns.colno,
            :syscolumns.coltype,
            :syscolumns.collength,
	    :systables.npused,
	    :systables.tabtype,
	    :syscolumns.colmin,
	    :syscolumns.colmax,
	    :systables.nrows,
	    :xtdname indicator xtdind,
	    :xtdtype,
	    :xtddesc,
	    :dbspace;

        if (sqlca.sqlcode && sqlca.sqlcode != SQLNOTFOUND ) {
	    fprintf( stderr,
		     "Error fetching a1 cursor;Error=%d\n",
		     sqlca.sqlcode);
	    rexit(1);
        } else {
	    if ((a_space=strchr( systables.tabname, ' ')) != 0)
		a_space[0] = '\0';
	    if ((sqlca.sqlcode && sqlca.sqlcode == SQLNOTFOUND)
		|| strcmp( prev.tabname, systables.tabname ) != 0)
	    {
		ucase( upname, systables.tabname );
		if (!pass_no) {
		    if (prev.tabtype[0] == 'T' || prev.tabtype[0] == 'V' || prev.tabtype[0] == 'E' || prev.tabtype[0] == 'S') {
			switch (OutType) {
			case FORTRAN:
			    fprintf( outfile,
				     "      end structure\n" );
			    fprintf( outfile,
				     "      record/%s_t/ %s\n",
				     prev.tabname,
				     prev.tabname );
			    fprintf( outfile,
				     "      common/%s_c/ %s\n",
				     prev.tabname,
				     prev.tabname );
			    break;
			case FourGL:
			    fprintf( outfile,
				     "\nEND RECORD\n" );
			case DDL:
			case ROW:
			    fprintf( outfile,
				     "\n);\n" );
			    break;
			case Listing:
			    fprintf( outfile, "\n" );
			    break;
			case CLang:
			case CPlus:
			    fprintf( outfile,
				     "} %s_t;\n",
				     prev.tabname );
			    break;
			}
		    }
		} else {
		    pass_no = 0;
		}
		i = 0;
		if (sqlca.sqlcode && sqlca.sqlcode == SQLNOTFOUND)
		    break;
		    
		if (systables.tabtype[0] == 'T'
		    || systables.tabtype[0] == 'V'
		    || systables.tabtype[0] == 'E'
		    || systables.tabtype[0] == 'S') {
		    if (multifile == 1) {
			if (outfile != (FILE *)NULL)
			    fclose( outfile );
			snprintf( filename, sizeof filename, "%s.%s.%s", databasename, systables.tabname, extension );
			outfile = fopen( filename, "w" );
			if (!outfile) {
			    fprintf( stderr, "Cannot open output: %s.\n", filename );
			    exit(3);
			}
		    }
		    switch (OutType) {
		    case FORTRAN:
			fprintf( outfile,
				 "\n      structure/%s_t/\n",
				 upname );
			break;
		    case FourGL:
			fprintf( outfile,
				 "\nDEFINE %s_rec RECORD\n",
				 systables.tabname );
			break;
		    case DDL:
			fprintf( outfile,
				 "\nCREATE TABLE %s (\n",
				 systables.tabname );
			break;
		    case ROW:
			fprintf( outfile,
				 "\nCREATE ROW TYPE %s_row (\n",
				 systables.tabname );
			break;
		    case CLang:
			fprintf( outfile,
				 "\ntypedef struct %s_S {\n",
				 upname );
			break;
		    case CPlus:
			fprintf( outfile,
				 "\nclass %s_C {\n",
				 upname );
			break;
		    }
		}
	    } else if (syscolumns.colno != 1 && (OutType == FourGL || OutType == DDL) || OutType == ROW) {
		fprintf( outfile, ",\n" );
	    } else if (syscolumns.colno != 1 && OutType == Listing) {
		fprintf( outfile, ", " );
	    }
	}
	
	if (!(systables.tabtype[0] == 'T' || systables.tabtype[0] == 'V' || systables.tabtype[0] == 'E' || systables.tabtype[0] == 'S')) {
	    memcpy( &prev, &systables, sizeof prev );
	    continue;
        }
	strtok( syscolumns.colname, " " );
        syscolumns.coltype &= SQLTYPE;
	if (OutType == Listing) {
	    fprintf( outfile, "%s", syscolumns.colname );
	} else if (OutType == DDL || OutType == ROW) {
	    fprintf( outfile,
		     "\t%s %s",
		     syscolumns.colname,
		     typename( syscolumns.coltype, syscolumns.collength, xtdname, xtdtype, xtddesc, dbspace ) );
	} else if (syscolumns.coltype == SQLCHAR) {
	    switch( OutType ) {
	    case FORTRAN:
		fprintf( outfile, "         character*%d %s\n",
			 syscolumns.collength,
			 syscolumns.colname );
		break;
	    case ESQL:
		if (Do_VCHAR) {
		    fprintf( outfile, "\tstring %s[%d];\n",
			     syscolumns.colname,
			     syscolumns.collength + 1);
		} else {
		    fprintf( outfile, "\tfixchar %s[%d];\n",
			     syscolumns.colname,
			     syscolumns.collength );
		}
		break;
	    case FourGL:
	    case DDL:
	    case ROW:
		fprintf( outfile,
			 "%s%s CHAR(%d)",
			 (OutType == FourGL) ? "    " : "\t",
			 syscolumns.colname,
			 syscolumns.collength );
		break;
	    case CLang:
	    case CPlus:
		fprintf( outfile, "\tchar %s[%d];\n",
			 syscolumns.colname,
			 syscolumns.collength );
		break;
	    }
	} else if (syscolumns.coltype == SQLUDTFIXED) {
	    switch( OutType ) {
	    case FourGL:
	        fprintf( outfile,
			 "    Type not implemented: %s %s(%d)",
			 syscolumns.colname,
			 xtdname,
			 syscolumns.collength );
		break;
	    case FORTRAN:
	        fprintf( outfile,
			 "         character*%d %s\n",
			 syscolumns.collength,
			 syscolumns.colname );
		break;
	    case ESQL:
	    	fprintf( outfile,
			 "\tstring %s[%d];  /* UDT Fixed: %s */\n",
			 syscolumns.colname,
			 syscolumns.collength + 1,
			 xtdname );
		break;
	    case CLang:
	    case CPlus:
	    default:
	    	fprintf( outfile,
			 "\tchar %s[%d];  /* UDT Fixed: %s */\n",
			 syscolumns.colname,
			 syscolumns.collength + 1,
			 xtdname );
		break;
	    }
	} else if ( syscolumns.coltype == SQLNCHAR ) {
	    switch( OutType ) {
	    case FourGL:
		fprintf( outfile,
			 "%s%s NCHAR(%d)",
			 (OutType == FourGL) ? "    " : "\t",
			 syscolumns.colname,
			 syscolumns.collength );
		break;
	    default:
		fprintf( outfile, "/* NCHAR TYPE(%d) for COLUMN (%s) not implemented */\n",
			 syscolumns.coltype, syscolumns.colname );
		break;
	    }
	} else if ( syscolumns.coltype == SQLNVCHAR ) {
	    switch( OutType ) {
	    case FourGL:
		fprintf( outfile,
			 "%s%s NVARCHAR(%d)",
			 (OutType == FourGL) ? "    " : "\t",
			 syscolumns.colname,
			 syscolumns.collength );
		break;
	    default:
		fprintf( outfile, "/* NVCHAR TYPE(%d) for COLUMN (%s) not implemented */\n",
			 syscolumns.coltype, syscolumns.colname );
		break;
	    }
        } else if ( syscolumns.coltype == SQLSMINT ) {
	    switch( OutType ) {
	    case FORTRAN:
		fprintf( outfile,
			 "         integer*2 %s\n",
			 syscolumns.colname );
		break;
	    case FourGL:
		fprintf( outfile,
		 	 "%s%s SMALLINT",
			 (OutType == FourGL) ? "    " : "\t",
			 syscolumns.colname );
		break;
	    default:
		fprintf( outfile, "\tshort %s;\n", syscolumns.colname );
		break;
	    }
        } else if ( syscolumns.coltype == SQLINT ) {
	    switch( OutType ) {
	    case FORTRAN:
		fprintf( outfile,
			 "         integer*4 %s\n",
			 syscolumns.colname );
		break;
	    case FourGL:
		fprintf( outfile,
			 "%s%s INT",
			 (OutType == FourGL) ? "    " : "\t",
			 syscolumns.colname );
		break;
	    default:
		fprintf( outfile, "\tint %s;\n", syscolumns.colname );
		break;
	    }
        } else if ( syscolumns.coltype == SQLINT8 ) {
	    switch( OutType ) {
	    case FORTRAN:
		fprintf( outfile,
			 "         integer*8 %s\n",
			 syscolumns.colname );
		break;
	    case ESQL:
		fprintf( outfile, "\tbigint %s;\n", syscolumns.colname );
		break;
	    case FourGL:
		fprintf( outfile, "    %s DECIMAL(20,0)",
			 syscolumns.colname );
		break;
	    default:
		fprintf( outfile, "\tlong long %s;\n", syscolumns.colname );
		break;
	    }
        } else if ( syscolumns.coltype == SQLINFXBIGINT ) {
	    switch( OutType ) {	    
	    case FORTRAN:
		fprintf( outfile,
			 "         integer*8 %s\n",
			 syscolumns.colname );
		break;
	    case ESQL:
		fprintf( outfile, "\tbigint %s;\n", syscolumns.colname );
		break;
	    case FourGL:
		fprintf( outfile, "    %s DECIMAL(20,0)",
			 syscolumns.colname );
		break;
	    default:
		fprintf( outfile, "\tlong long %s;\n", syscolumns.colname );
		break;
	    }
        } else if ( syscolumns.coltype == SQLFLOAT) {
	    switch( OutType ) {
	    case FORTRAN:
		fprintf( outfile,
			 "         real*8 %s\n",
			 syscolumns.colname );
		break;
	    case FourGL:
		fprintf( outfile,
			 "%s%s FLOAT",
			 (OutType == FourGL) ? "    " : "\t",
			 syscolumns.colname );
		break;
	    default:
		fprintf( outfile, "\tdouble %s;\n", syscolumns.colname );
		break;
	    }
        } else if ( syscolumns.coltype == SQLSMFLOAT) {
	    switch( OutType ) {
	    case FORTRAN:
		fprintf( outfile,
			 "         real*4 %s\n",
			 syscolumns.colname );
		break;
	    case FourGL:
		fprintf( outfile,
			 "%s%s SMALLFLOAT",
			 (OutType == FourGL) ? "    " : "\t",
			 syscolumns.colname );
		break;
	    default:
		fprintf( outfile, "\tfloat %s;\n", syscolumns.colname );
		break;
	    }
	} else if ( syscolumns.coltype == SQLDECIMAL) {
	    switch( OutType ) {
	    case FORTRAN:
		fprintf( outfile,
			 "C        record/dec_t/ %s !NOT SUPPORTED YET\n",
			 syscolumns.colname );
		break;
	    case FourGL:
		if (PRECDEC(syscolumns.collength) == 255)
		    fprintf( outfile,
			     "%s%s DECIMAL(%d)",
			     (OutType == FourGL) ? "    " : "\t",
			     syscolumns.colname,
			     PRECTOT(syscolumns.collength) );
		else
		    fprintf( outfile,
			     "%s%s DECIMAL(%d,%d)",
			     (OutType == FourGL) ? "    " : "\t",
			     syscolumns.colname,
			     PRECTOT(syscolumns.collength),
			     PRECDEC(syscolumns.collength) );
		break;
	    default:
		fprintf( outfile, "\tdec_t %s;\n", syscolumns.colname );
		break;
	    }
        } else if ( syscolumns.coltype == SQLSERIAL) {
	    switch( OutType ) {
	    case FORTRAN:
		fprintf( outfile,
			 "         integer*4 %s;  ! SERIAL\n",
			 syscolumns.colname );
		break;
	    case FourGL:
		fprintf( outfile,
			 "%s%s INT",
			 (OutType == FourGL) ? "    " : "\t",
			 syscolumns.colname );
		break;
	    default:
		fprintf( outfile,
			 "\tint %s;  /* SERIAL */\n",
			 syscolumns.colname );
		break;
	    }
        } else if ( syscolumns.coltype == SQLSERIAL8) {
	    switch( OutType ) {
	    case FORTRAN:
		fprintf( outfile,
			 "         integer*8 %s;  ! SERIAL8\n",
			 syscolumns.colname );
		break;
	    case FourGL:
		fprintf( outfile,
			 "%s%s DECIMAL(20,0)",
			 (OutType == FourGL) ? "    " : "\t",
			 syscolumns.colname );
		break;
	    case ESQL:
		fprintf( outfile,
			 "\tbigint %s;  /* SERIAL8 */\n",
			 syscolumns.colname );
		break;
	    default:
		fprintf( outfile,
			 "\tlong long %s;  /* SERIAL8 */\n",
			 syscolumns.colname );
		break;
	    }
        } else if ( syscolumns.coltype == SQLBIGSERIAL) {
	    switch( OutType ) {
	    case FORTRAN:
		fprintf( outfile,
			 "         integer*8 %s;  ! BIGSERIAL\n",
			 syscolumns.colname );
		break;
	    case FourGL:
		fprintf( outfile,
			 "    %s DECIMAL(20,0)",
			 syscolumns.colname );
		break;
	    case ESQL:
		fprintf( outfile,
			 "\tbigint %s;  /* BIGSERIAL */\n",
			 syscolumns.colname );
		break;
	    default:
		fprintf( outfile,
			 "\tlong long %s;  /* BIGSERIAL */\n",
			 syscolumns.colname );
		break;
	    }
        } else if ( syscolumns.coltype == SQLDATE) {
	    switch( OutType ) {
	    case FORTRAN:
		fprintf( outfile,
			 "         integer*4 %s  ! INFORMIX DATE\n",
			 syscolumns.colname );
		break;
	    case FourGL:
		fprintf( outfile,
			 "%s%s DATE",
			 (OutType == FourGL) ? "    " : "\t",
			 syscolumns.colname );
		break;
	    default:
		fprintf( outfile, "\ttime_t %s;  /* DATE */\n", syscolumns.colname );
		break;
	    }
        } else if ( syscolumns.coltype == SQLMONEY) {
	    switch( OutType ) {
	    case FORTRAN:
		fprintf( outfile,
			 "C        record/dec_t/ %s   ! MONEY - NOT YET SUPPORTED\n",
			 syscolumns.colname );
		break;
	    case FourGL:
		if (PRECDEC(syscolumns.collength) == 255)
		    fprintf( outfile,
			     "%s%s MONEY(%d)",
			     (OutType == FourGL) ? "    " : "\t",
			     syscolumns.colname,
			     PRECTOT(syscolumns.collength) );
		else
		    fprintf( outfile,
			     "%s%s MONEY(%d,%d)",
			     (OutType == FourGL) ? "    " : "\t",
			     syscolumns.colname,
			     PRECTOT(syscolumns.collength),
			     PRECDEC(syscolumns.collength) );
		break;
	    default:
		fprintf( outfile,
			 "\tdec_t %s;  /* MONEY */\n",
			 syscolumns.colname );
		break;
	    }
        } else if ( syscolumns.coltype == SQLDTIME) {
	    switch( OutType ) {
	    case FORTRAN:
		fprintf( outfile,
			 "C        record/dtime_t/ %s  "
			 "! DATETIME - NOT YET SUPPORTED\n",
			 syscolumns.colname );
		break;
	    case FourGL:
		fprintf( outfile, "    %s DATETIME ", syscolumns.colname );

		switch (TU_START( syscolumns.collength )) {
		case TU_YEAR:
		    fprintf( outfile, "YEAR" );
		    break;
		case TU_MONTH:
		    fprintf( outfile, "MONTH" );
		    break;
		case TU_DAY:
		    fprintf( outfile, "DAY" );
		    break;
		case TU_HOUR:
		    fprintf( outfile, "HOUR" );
		    break;
		case TU_MINUTE:
		    fprintf( outfile, "MINUTE" );
		    break;
		case TU_SECOND:
		    fprintf( outfile, "SECOND" );
		    break;
		}
		fprintf( outfile, " TO " );
		switch (TU_END( syscolumns.collength )) {
		case TU_MONTH:
		    fprintf( outfile, "MONTH" );
		    break;
		case TU_DAY:
		    fprintf( outfile, "DAY" );
		    break;
		case TU_HOUR:
		    fprintf( outfile, "HOUR" );
		    break;
		case TU_MINUTE:
		    fprintf( outfile, "MINUTE" );
		    break;
		case TU_SECOND:
		    fprintf( outfile, "SECOND" );
		    break;
		case TU_F1:
		    fprintf( outfile, "FRACTION(1)" );
		    break;
		case TU_F2:
		    fprintf( outfile, "FRACTION(2)" );
		    break;
		case TU_F3:
		    fprintf( outfile, "FRACTION(3)" );
		    break;
		case TU_F4:
		    fprintf( outfile, "FRACTION(4)" );
		    break;
		case TU_F5:
		    fprintf( outfile, "FRACTION(5)" );
		    break;
		}
		break;
	    default:
		fprintf( outfile,
			 "\tdtime_t %s;  /* DATETIME */\n",
			 syscolumns.colname );
		break;
	    }
        } else if ( syscolumns.coltype == SQLINTERVAL) {
	    switch( OutType ) {
	    case FORTRAN:
		fprintf( outfile,
			 "C        record/intrvl_t/ %s   "
			 "! INTERVAL - NOT YET SUPPORTED\n",
			 syscolumns.colname );
		break;
	    case FourGL:
		fprintf( outfile, "    %s INTERVAL ", syscolumns.colname );
		switch (TU_START( syscolumns.collength )) {
		case TU_YEAR:
		    fprintf( outfile,
			     "YEAR(%d)",
			     TU_LEN( syscolumns.collength ) + TU_START( syscolumns.collength ) - TU_END( syscolumns.collength ) );
		    break;
		case TU_MONTH:
		    fprintf( outfile,
			     "MONTH(%d)",
			     TU_LEN( syscolumns.collength ) + TU_START( syscolumns.collength ) - TU_END( syscolumns.collength ) );
		    break;
		case TU_DAY:
		    fprintf( outfile,
			     "DAY(%d)",
			     TU_LEN( syscolumns.collength ) + TU_START( syscolumns.collength ) - TU_END( syscolumns.collength ) );
		    break;
		case TU_HOUR:
		    fprintf( outfile,
			     "HOUR(%d)",
			     TU_LEN( syscolumns.collength ) + TU_START( syscolumns.collength ) - TU_END( syscolumns.collength ) );
		    break;
		case TU_MINUTE:
		    fprintf( outfile,
			     "MINUTE(%d)",
			     TU_LEN( syscolumns.collength ) + TU_START( syscolumns.collength ) - TU_END( syscolumns.collength ) );
		    break;
		case TU_SECOND:
		    fprintf( outfile,
			     "SECOND(%d)",
			     TU_LEN( syscolumns.collength ) + TU_START( syscolumns.collength ) - TU_END( syscolumns.collength ) );
		    break;
		}
		fprintf( outfile, " TO " );
		switch (TU_END( syscolumns.collength )) {
		case TU_MONTH:
		    fprintf( outfile, "MONTH" );
		    break;
		case TU_DAY:
		    fprintf( outfile, "DAY" );
		    break;
		case TU_HOUR:
		    fprintf( outfile, "HOUR" );
		    break;
		case TU_MINUTE:
		    fprintf( outfile, "MINUTE" );
		    break;
		case TU_SECOND:
		    fprintf( outfile, "SECOND" );
		    break;
		case TU_F1:
		    fprintf( outfile, "FRACTION(1)" );
		    break;
		case TU_F2:
		    fprintf( outfile, "FRACTION(2)" );
		    break;
		case TU_F3:
		    fprintf( outfile, "FRACTION(3)" );
		    break;
		case TU_F4:
		    fprintf( outfile, "FRACTION(4)" );
		    break;
		case TU_F5:
		    fprintf( outfile, "FRACTION(5)" );
		    break;
		}
		break;
	    default:
                fprintf( outfile,
			 "\tintrvl_t %s;  /* INTERVAL */\n",
			 syscolumns.colname );
		break;
	    }
	} else if ( syscolumns.coltype == SQLVCHAR) {
	    switch( OutType ) {
	    case FORTRAN:
		fprintf( outfile,
			 "         character*%d %s\n",
			 syscolumns.collength % 256,
			 syscolumns.colname );
		break;
	    case ESQL:
		fprintf( outfile,
			 "\tstring %s[%d];\n",
			 syscolumns.colname,
			 (syscolumns.collength % 256) + 1 );
		break;
	    case FourGL:
		fprintf( outfile, "    %s VARCHAR(%d)", syscolumns.colname, syscolumns.collength );
		break;
	    default:
		fprintf( outfile,
			 "\tchar %s[%d];\n",
			 syscolumns.colname,
			 (syscolumns.collength % 256) + 1 );
		break;
	    }
	} else if ( syscolumns.coltype == SQLUDTVAR) {
	    switch( OutType ) {
	    case FORTRAN:
		fprintf( outfile,
			 "         character*%d %s  ! %s\n",
			 syscolumns.collength,
			 syscolumns.colname,
			 xtdname );
		break;
	    case ESQL:
		fprintf( outfile,
			 "\tstring %s[%d];  /* UDT Variable: %s */\n",
			 syscolumns.colname,
			 syscolumns.collength + 1,
			 xtdname );
		break;
	    case FourGL:
		fprintf( outfile,
			 "    %s CHAR(%d)",
			 syscolumns.colname,
			 syscolumns.collength );
		break;
	    default:
		fprintf( outfile,
			 "\tchar %s[%d];  /* UDT Variable: %s */\n",
			 syscolumns.colname,
			 syscolumns.collength + 1,
			 xtdname );
		break;
	    }
	} else if ( syscolumns.coltype == SQLBYTES || syscolumns.coltype == SQLTEXT) {
	    switch( OutType ) {
	    case FORTRAN:
		fprintf( outfile,
			 "C        BLOB Fields not supported by FORTRAN.\n" );
		break;
	    case FourGL:
		fprintf( outfile,
			 "    loc_t %s",
			 syscolumns.colname );
		break;
	    default:
		fprintf( outfile,
			 "\tloc_t %s;\n",
			 syscolumns.colname );
		break;
	    }
	} else {
	    switch( OutType ) {
	    case FORTRAN:
		fprintf( outfile,
			 "%s %s\n",
			 typename_f( syscolumns.coltype, syscolumns.collength, xtdname, xtdtype, xtddesc, dbspace ),
			 syscolumns.colname );
		break;
	    case FourGL:
		fprintf( outfile,
			 "    %s %s",
			 syscolumns.colname,
			 typename( syscolumns.coltype, syscolumns.collength, xtdname, xtdtype, xtddesc, dbspace ) );
		break;
	    default:
		fprintf( outfile,
			 "\t%s %s;\n",
			 syscolumns.colname,
			 typename( syscolumns.coltype, syscolumns.collength, xtdname, xtdtype, xtddesc, dbspace ) );
		break;
	    }
	}
	memcpy( &prev, &systables, sizeof prev );
    }
    if (OutType == ESQL)
	fprintf( outfile, "EXEC SQL END DECLARE SECTION;\n" );
    }

void rexit(int stat)
{

    if (stat ) {
        EXEC SQL
	    close a1;
        if (sqlca.sqlcode) {
            fprintf( stderr,
		     "cannot close a1 cursor;Error=%d\n",
		     sqlca.sqlcode);
        }
    }
    EXEC SQL
	CLOSE DATABASE;
    EXEC SQL
	DISCONNECT ALL;
    exit(stat);
}

void usage( char *program )
{
    fprintf( stderr,
	     "Usage:\n\t%s [-F|G|P|R|D|e|f|l] [-h host] -d database [-t table] [filename]\n",
	     basename( program ) );
    fprintf( stderr,
	     "\t%s -M [-F|G|D|e|f|l] [-h host] -d database [-t table] [extension]\n",
	     basename( program ) );
    fprintf( stderr,
	     "\t%s -V\n", basename( program ) );
    fprintf( stderr,
	     "\t-V - Display revision and copyright and exit.\n" );
    fprintf( stderr,
	     "\t-F - generate FORTRAN STRUCTURES rather than \"C\".\n" );
    fprintf( stderr,
	     "\t-G - generate 4GL RECORDs rather than \"C\".\n" );
    fprintf( stderr,
	     "\t-e - generate ESQL-C style include file.\n" );
    fprintf( stderr,
	     "\t-P - generate a C++ class structure.\n" );
    fprintf( stderr,
	     "\t-f - generate 'fixchar' type for strings in ESQL headers rather than\n"
	     "\t    'string' type. Implies -e.\n" );
    fprintf( stderr,
	     "\t-D - generate simple SQL DDL for the tables.\n" );
    fprintf( stderr,
	     "\t-R - generate SQL DDL for a ROW type.\n" );
    fprintf( stderr,
	     "\t-l - generate a simple comma separated column name list suitable for \n"
	     "\t     use in projection clauses or code.\n" );
    fprintf( stderr,
	     "\t-M - Write each table's output to a file named <database>.<table>.<ext>\n"
	     "\t     where <ext> defaults to an appropriate extension for the type of\n"
	     "\t     structures being generated (.c, .ec, .4gl, .f77, .sql). If -M is\n"
	     "\t     passed the optional extension argument overrides the default file\n"
	     "\t     extension.\n" );
    fprintf( stderr,
	     "\t-r - ignore REPLCHECK shadow columns.\n" );
    fprintf( stderr,
	     "\t-v - ignore VERCOLS shadow columns.\n" );
    fprintf( stderr,
	     "\t-k - ignore ERKEY shadow columns.\n" );
    fprintf( stderr, "\tHost defaults to current host.\n" );
    fprintf( stderr,
	     "\tTable may contain valid \"matches\" clause wildcards.\n" );
    fprintf( stderr,
	     "\tTable defaults to all user tables %s\n\n",
	     "(excluding system catalog tables.)" );
    fprintf( stderr, "\tOutput format defaults to C/C++ struct definitions.\n" );
    fprintf( stderr,
	     "\tIf table contains columns of a type not supported by the output format\n"
	     "\tthe line is produced as a comment\n" );
}

void ucase( char *target, char *source )
{
    while (*source)
	*target++ = toupper( *source++ );
    *target = (char)0;
}

char *typename( int coltype, int collength, char *xtdname, int xtdtype, char *xtddesc, char *dbspace )
{
    int p1, p2, pt1, pt2, fr1, fr2;
    char part1[20], part2[20];
    static char name[500];

    switch( coltype ) {
    case SQLCHAR:
	snprintf( name, sizeof name, "CHAR(%d)", collength );
	break;
    case SQLNCHAR:
	snprintf( name, sizeof name, "NCHAR(%d)", collength );
	break;
    case SQLSET:
    case SQLMULTISET:
    case SQLLIST:
	snprintf( name, sizeof name, "%s", xtddesc );
	break;
    case SQLINFXBIGINT:
	snprintf( name, sizeof name, "%s", "BIGINT" );
	break;
    case SQLINT8:
	snprintf( name, sizeof name, "%s", "INT8" );
	break;
    case SQLINT:
	snprintf( name, sizeof name, "%s", "INT" );
	break;
    case SQLSMINT:
	snprintf( name, sizeof name, "%s", "SMALLINT" );
	break;
    case SQLFLOAT:
	snprintf( name, sizeof name, "%s", "FLOAT" );
	break;
    case SQLSMFLOAT:
	snprintf( name, sizeof name, "%s", "SMALLFLOAT" );
	break;
    case SQLSERIAL:
	snprintf( name, sizeof name, "%s", "SERIAL" );
	break;
    case SQLSERIAL8:
	snprintf( name, sizeof name, "%s", "SERIAL8" );
	break;
    case SQLBIGSERIAL:
	snprintf( name, sizeof name, "%s", "BIGSERIAL" );
	break;
    case SQLDECIMAL:
	if ((collength % 256) == 255)
	    snprintf( name, sizeof name, "DECIMAL(%d)", (int)(collength / 256) );
	else
	    snprintf( name, sizeof name, "DECIMAL(%d,%.1d)", (int)(collength / 256), (int)(collength % 256) );
	break;
   case SQLUDTVAR:
    case SQLUDTFIXED:
    case SQLROW:
	switch (xtdtype) {
	case 15:
	case 16:
	case 17:
	case 24:
	case 25:
	case 26:
	    snprintf( name, sizeof name, "%s", xtdname );
	    break;
	default:
	    snprintf( name, sizeof name, "%s(%d)", xtdname, collength );
	    break;
	}
	break;
    case SQLDATE:
	snprintf( name, sizeof name, "%s", "DATE" );
	break;
    case SQLDTIME:
        pt1 = (((collength / 16) % 16) <= 10) ?
            (((collength / 16) % 16) / 2) : 6;
        strcpy( part1, dtparts[pt1] );
        pt2 = ((collength % 16) <= 10) ?
            ((collength % 16) / 2) : 6;
        if (pt2 == 6) {
            fr2 = (collength % 16) - 10;
            snprintf( part2, sizeof part2, "%s(%2d)", dtparts[pt2], fr2 );
        } else
            strcpy( part2, dtparts[pt2] );
        snprintf( name,
		  sizeof name,
		  "DATETIME %s TO %s",
		  part1,
		  part2 );
	break;
    case SQLINTERVAL:
        p1 = (collength / 16) % 16;
        pt1 = (p1 <= 10) ? (p1 / 2) : 6;
        fr2 = (collength % 16) - 10;
        p2 = collength % 16;
        pt2 = (p2 <= 10) ? (p2 / 2) : 6;
        if (pt1 != 6) {
            fr1 = (collength / 256) + p1 - p2;
            snprintf( part1, sizeof part1, "%s(%d)", dtparts[pt1], fr1 );
        } else
            strcpy( part1, dtparts[pt1] );
        if (pt2 == 6) {
            snprintf( part2, sizeof part2, "%s(%d)", dtparts[pt2], fr2 );
        } else
            strcpy( part2, dtparts[pt2] );
        snprintf( name,
		  sizeof name,
		  "INTERVAL %s TO %s",
		  part1,
		  part2 );
	break;
    case SQLVCHAR:
	snprintf( name,
		  sizeof name,
		  "VARCHAR(%u,%.1u)",
		  (unsigned)collength % (unsigned)256,
		  ((unsigned)collength / (unsigned)256) % (unsigned)256 );
	break;
    case SQLNVCHAR:
	snprintf( name,
		  sizeof name,
		  "NVARCHAR(%u,%.1u)",
		  (unsigned)collength % (unsigned)256,
		  ((unsigned)collength / (unsigned)256) % (unsigned)256 );
	break;
    case SQLBYTES:
	snprintf( name, sizeof name, "BYTE IN %s", dbspace );
    break;
    case SQLTEXT:
	snprintf( name, sizeof name, "TEXT IN %s", dbspace );
    break;
    }
    return name;
}



char *typename_f( int coltype, int collength, char *xtdname, int xtdtype, char *xtddesc, char *dbspace )
{
    int p1, p2, pt1, pt2, fr1, fr2, ix, cnt;
    char part1[10], part2[10];
    static char name[500];

    switch( coltype ) {
    case SQLCHAR:
    case SQLNCHAR:
	snprintf( name, sizeof name, "        character*%d", collength );
	break;
    case SQLSET:
    case SQLMULTISET:
    case SQLLIST:
	snprintf( name, sizeof name, "        %s", xtddesc );
	break;
    case SQLINFXBIGINT:
    case SQLINT8:
	snprintf( name, sizeof name, "        %s", "int*8" );
	break;
    case SQLINT:
	snprintf( name, sizeof name, "        %s", "int*4" );
	break;
    case SQLSMINT:
	snprintf( name, sizeof name, "        %s", "int*2" );
	break;
    case SQLFLOAT:
    case SQLDECIMAL:
	snprintf( name, sizeof name, "        %s", "double" );
	break;
    case SQLSMFLOAT:
	snprintf( name, sizeof name, "        %s", "float" );
	break;
    case SQLSERIAL:
	snprintf( name, sizeof name, "        %s", "int*4" );
	break;
    case SQLSERIAL8:
    case SQLBIGSERIAL:
	snprintf( name, sizeof name, "        %s", "int*8" );
	break;
    case SQLUDTVAR:
    case SQLUDTFIXED:
    case SQLROW:
	switch (xtdtype) {
	case 15:
	case 16:
	case 17:
	case 24:
	case 25:
	case 26:
	    snprintf( name, sizeof name, "C       NO EQUIVALENT TYPE: %s", xtdname );
	    break;
	default:
	    snprintf( name, sizeof name, "C       NO EQUIVALENT TYPE: %s(%d)", xtdname, collength );
	    break;
	}
	break;
    case SQLDATE:
	snprintf( name, sizeof name, "        %s", "int*4" );
	break;
    case SQLDTIME:
        pt1 = (((collength / 16) % 16) <= 10) ?
            (((collength / 16) % 16) / 2) : 6;
        strcpy( part1, dtparts[pt1] );
        pt2 = ((collength % 16) <= 10) ?
            ((collength % 16) / 2) : 6;
        if (pt2 == 6) {
            fr2 = (collength % 16) - 10;
            snprintf( part2, sizeof part2, "%s(%d)", dtparts[pt2], fr2 );
        } else
            strcpy( part2, dtparts[pt2] );
	for (cnt=0, ix=pt1; ix++; ix <= pt2) {
	    cnt += ptcnt[ix];
	}
	if (pt2 == 6) cnt += fr2;
	snprintf( name, sizeof name, "        character*%d", cnt );
	break;
    case SQLINTERVAL:
        snprintf( name, sizeof name, "C       NO EQUIVALENT TYPE: INTERVAL" );
	break;
    case SQLVCHAR:
    case SQLNVCHAR:
	snprintf( name,
		  sizeof name,
		  "CHARACTER*%u",
		  (unsigned)collength % (unsigned)256 );
	break;
    case SQLBYTES:
	snprintf( name, sizeof name, "C       NO EQUIVALENT TYPE: BYTE" );
    break;
    case SQLTEXT:
	snprintf( name, sizeof name, "C       NO EQUIVALENT TYPE: TEXT" );
    break;
    }
    return name;
}
