/* print_indexes.ec - Print out indexes for a table.
*/
/*
   RCS Header:
   -----------
   $Header: /home/art/GoogleDrive/Latest\040Art's\040Stuff/myschema.d/fork/RCS/print_indexes.ec,v 2.77 2023/12/11 14:21:26 art Exp $

   RCS Log:
   --------
   $Log: print_indexes.ec,v $
   Revision 2.77  2023/12/11 14:21:26  art
   Added support for the LARGE and SMALL keywords in CREATE TABLE
   statements.

   Revision 2.76  2022/01/23 15:59:47  art
   Fixed problem that only printed the first partition of indexes partitioned by expression.

   Revision 2.75  2021/05/20 22:03:24  art
   Fixed printout of the traditional replcheck index when it
   includes the erkey columns. Was being suppressed like the
   automatic erkey index.

   Revision 2.74  2021/03/09 16:19:40  art
   New option --for-er adds WITH REPLCHECK and WITH ERKEY to each table to 
   create a schema suitable for building an ER ready copy of the database.
   Also generates code to create the index on the primary key column and
   the ifx_replcheck column if a primary key exists.

   Revision 2.73  2019/09/13 15:46:24  art
   Added support for the v14.10.FC2 partial index feature!

   Revision 2.72  2018/12/31 22:04:29  art
   Making changes to get myschema to compile/link under GCC v7+

   Revision 2.71  2015/11/24 17:09:46  art
   Fixed routine owner handling where there are "--" comments embedded in the parameter list.
   Fixed the column list included in UPDATE STATISTICS LOW when -u is passed.
   Fixed an issue with the DROP statement not printing ROW type parameters when --drop-proc is passed.
   Thanks to David Link and Snorri Bergmann for pointing these out.

   Revision 2.70  2015/07/07 14:44:58  art
   Typo.

   Revision 2.69  2015/07/07 14:43:11  art
   Removed support for spurious BSON functions that were mistakenly documented but don't exist.

   Revision 2.68  2015/07/07 14:29:45  art
   Added some diagnostics.  Protected extended type code from being executed on a v7 server.

   Revision 2.67  2015/04/14 19:10:50  art
   Fixed handling of functional indexes where the index has one or more constant arguments.

   Revision 2.66  2015/04/14 18:14:26  art
   Fixed handling of BSON conversion functional indexes (using the bson_value_*() functions).
   All except the new bson_value_float() function 'cause I don't have a copy of v12.10.xC5
   to get the "special" procid for that function from.  I do wish IBM would stop breaking
   their own protocols and had registered these functions in sysprocedures like good
   boys and girls!

   Revision 2.65  2015/04/14 14:35:47  art
   Fixed indexes against some BSON function types and for any indexes containing BSON columns.

   Revision 2.64  2014/11/07 12:09:59  art
   Fixed a problem causing two indexes with the same key to be generated when a
   has a unique constraint and one of two or more foreign keys using the same
   key.

   Revision 2.63  2014/10/21 03:23:25  art
   Missing a criteria for 2.62 fix.

   Revision 2.62  2014/10/21 03:18:20  art
   Fixed a problem I found at one client. I have found orphaned indexes.
   Disabled indexes with a generated name that were created by a constraint,
   but the constraint was no longer in existence.

   Revision 2.61  2014/02/03 14:48:47  art
   Arg.  Fixed a bug that causes some indexes to not print.
   Thanks to Mike Walker!

   Revision 2.60  2013/12/29 20:03:19  art
   Added support for v12.10 ROLLING fragments DETACH and DISCARD keywords.
   Support for DETACH (ANY|INTERVAL FIRST|INTERVAL ONLY) no in there yet
   because the documented syntax doesn't work!

   Revision 2.59  2013/11/01 12:00:10  art
   Fixed a bug that caused indexes on catalog tables to be skipped.

   Revision 2.58  2013/07/29 21:00:09  art
   typo.

   Revision 2.57  2013/07/29 20:57:30  art
   Fixed handling of the new flag for compressed indexes.  Was being included for any 11.70
   release, but it does not appear until one of the late maintenance releases.  Not needed
   until v12.10 for compressed indexes.

   Revision 2.56  2013/03/19 23:12:54  art
   Fixed handling of attached/IN TABLE indexes. Fixed a problem when
   -O and --drop-proc are specified together causing the DROP routine
   statement to have no argument list. Added missing support for
   REFERENCES BYTE/TEXT arguments in DROP routine statements when
   --drop-proc is specified.

   Revision 2.55  2013/03/14 03:44:21  art
   Added new option --no-storage-options which suppresses "IN <dbspace>" and
   FRAGMENT BY clauses.  This can ease some migrations.

   Revision 2.54  2013/03/11 16:55:30  art
   Added new option -K to name autonamed indexes and unnamed constraints as <table>_pk
   for primary key indexes and <table>_<type>### where type is one of "uk", "fk", "ak"
   and ### is a sequential number within the type.
   Fixed -U.
   Enhanced -v to print progress after "Analyzing schema".
   reordered index printouts to group by the type of constraint each supports if any.

   Revision 2.53  2013/01/24 00:03:49  art
   Added Centaurus feature support.

   Revision 2.52  2012/06/08 21:12:28  art
   Removed messages about constraint indexes being stored in rootdb.  No longer true.

   Revision 2.51  2012/03/12 00:19:49  art
   Added prototype for local index state printout function.

   Revision 2.50  2012/03/12 00:16:51  art
   Added support for new option --index-online to add the ONLINE keyword to create index
   statements.  Added support for foreign keys created with the INDEX DISABLED option.

   Revision 2.49  2011/03/04 17:11:25  art
   Fixed other references to syscolumns.colattr that break myschema when run against
   versions earlier than 11.50.

   Revision 2.48  2010/12/08 14:57:32  art
   Fixed a typo in print_indexes.ec

   Revision 2.47  2010/12/06 16:28:04  art
    Fixed two warnings and an SQL bug.  Thanks to Oyvind Gjarstad

   Revision 2.46  2010/11/17 20:49:40  art
   Added support for WITH REPLCHECK, WITH ERKEY, and WITH AUDIT clauses in CREATE TABLE.

   Revision 2.45  2010/11/15 15:53:54  art
   Forest of Trees index and Index Extent Sizing support completed.

   Revision 2.44  2010/11/08 20:40:31  art
   Added new --create_conditionally feature to optionally add " IF NOT EXISTS " clauses to c
   create object statements.  Feature supported by 11.70.

   Revision 2.43  2010/08/25 13:52:14  art
   Fixed several compile time warnings.  Thanks to John Adamski for identifying them.

   Revision 2.42  2010/01/06 13:55:18  art
   Added new option --set-owner to override all object owners with a specified owner name.

   Revision 2.41  2009/11/23 18:47:05  art
   Added support for EXTERNAL tables (11.50xc6+).
   Improved 64bit compiler support.  Removed warnings.

   Revision 2.40  2009/10/02 06:24:56  art
   Added new option -g to place user privileges into a separate file.
   Improved trigger parser.
   Made indexes and constraint processing recognize and report object state (ie disabled,
   filtering, etc.).
   Improved --filtering option again.

   Revision 2.39  2009/02/11 18:30:34  art
   Fixed problem with LVARCHARS on AIX-64bit.

   Revision 2.38  2009/01/01 17:58:25  art
   Removed RCSWhat ident string which was BB specific.

   Revision 2.37  2008/12/30 17:43:39  art
   Fixed bad cast for null string terminator (NULL s/b 0)

   Revision 2.36  2007/06/01 18:54:35  kagel
   added named partition support for indexes.

   Revision 2.35  2004/10/18 19:10:35  kagel
   Fixed problem with ASC/DESC specifiers for non-BTREE indexes.

   Revision 2.34  2004/09/14 17:19:17  kagel
   Fixing ANSI database access.

   Revision 2.33  2004/08/18 18:15:08  kagel
   Fixed typo.

   Revision 2.32  2004/08/18 17:39:13  kagel
   Added new option, -k, to toggle the creation of named indexes to replace
   hidden constraint indexes.
   Increased features version to 6.29 to recognized the new feature.

   Revision 2.31  2004/08/04 17:59:54  kagel
   fixed missing presumed type.

   Revision 2.30  2004/06/22 14:16:33  kagel
   Added SET EXPLAIN to the sysmaster queries when -E is passed.
   Fixed missing 'IN dbspace' clause for constraint indexes which would
   cause a syntax error for PRIMARY KEY and UNIQUE constraints on a table
   fragmented ROUND ROBIN.  Thanks to Ravi Krishna for pointing this one out.

   Revision 2.29  2003/10/29 17:39:01  kagel
   Typo

   Revision 2.28  2003/10/29 15:47:24  kagel
   Improved backward compatibility by IFDEFing use of lvarchar host vars.

   Revision 2.27  2003/08/13 15:17:32  kagel
   Fixed 9.xx functional index key handling.  Thanks to Ravi Krishna for
   pointing out the problem and supplying a great example.  Correctly handles
   descending functional keys (dbschema does not as of 9.4UC2 at least).

   Revision 2.26  2003/07/30 18:27:13  kagel
   Made the first Kludge handle other unregistered access methods with a
   comment which identifies the am_id and will be a syntax error when the
   schema is run.

   Revision 2.25  2003/07/30 15:41:53  kagel
   Added two kludges to handle Excalibur Datablade indexes.  The DB does not
   properly define its access method (etx) with a row in sysams so have to
   Kludge it to make the access method "etx" when am_id = 100 and no sysams
   record is found.  The second Kludge to include the opclass name if there
   is no sysams record is generic and will work with any unknown access method.
   Also noted that the code was printing the opclass id rather than the name,
   fixed that while I was there.

   Revision 2.24  2003/01/10 17:17:59  kagel
   Fixed reference to sysopclasses.

   Revision 2.23  2003/01/09 20:29:11  kagel
   Fixed a typo in the main query affecting 9.xx users.

   Revision 2.22  2003/01/02 21:14:20  kagel
   Fixed formatting of index lines.

   Revision 2.21  2003/01/02 20:51:33  kagel
   Added code to handle functional indexes and non-default op=classes for
   indexed UDT columns.

   Revision 2.20  2002/11/20 21:47:54  kagel
   Added DELIMIDENT support, RAW table support.

   Revision 2.19  2002/11/06 16:03:29  kagel
   Corrected the Universal server handling of index keys.  The ikeyextractcolno
   function ONLY works for vanilla types.  Need to retrieve and decode the
   indexkeys column in code.  Copied new code from dostats, inserted a note
   indicating where op class support needs to be inserted.

   Revision 2.18  2001/10/31 20:00:03  kagel
   Made a change to match the output of dbschema.  For some simple indexes
   dbschema is outputting the access method even if it is just 'btree'.  I
   have changed print_indexes to always output the access method on servers
   that provide that information (ie 9.xx).

   Revision 2.17  2001/10/24 18:00:34  kagel
   Protected the #define of _H_LOCALEDEF from multiple definition.  This is
   needed for some versions of AIX but apparently later versions of the
   Informix SDK already define it.

   Revision 2.16  2001/09/28 16:42:12  kagel
   Fixed retro problem with 5/6/7/SE servers due to new code to 9.xx.

   Revision 2.15  2001/09/26 17:17:33  kagel
   Added Access Method support.
   Using sysindices instead of sysindexes in 9.xx.

   Revision 2.14  2001/09/14 17:45:39  kagel
   Object support added.

   Revision 2.13  2001/06/05 13:10:38  kagel
   Added ownership to indexes.

   Revision 2.12  2001/01/16 22:16:00  kagel
   Included errno.h.

   Revision 2.11  2001/01/16 21:55:31  kagel
   Added protection to malloc() calls.

   Revision 2.10  2000/07/11 17:12:13  kagel
   Added support for FILLFACTOR using an environment variable and new -m
   flag to force using minimal initial extent when current extent is too
   large rather than just printing a warning.

   Revision 2.9  2000/04/14 16:49:47  kagel
   Fixed ORDER BY clause to a better order though the actual output should not
   change.

   Revision 2.8  2000/04/12 18:29:03  kagel
   Fixed problem affecting implicitely named constraint indexes which are
   referenced by more than one constraint.  This caused multiple indexes with
   different explicit names to be generated with identical keys.  Indexes are
   now output in the following order: any clustered index first followed by
   other indexes in order by indexname.

   Revision 2.7  1999/10/26 22:21:43  kagel
   Fixed minor problems.

   Revision 2.6  1999/10/26 21:44:02  kagel
   typo revisited.

   Revision 2.5  1999/10/26 21:41:49  kagel
   typo.

   Revision 2.4  1999/10/26 21:23:45  kagel
   Added new feature to duplicate current levels of UPDATE STATISTICS.  By
   default written to filename3, optionally to a separate file.
   Modified function to output the LOW stats statements for each index.

   Revision 2.3  1999/10/12 16:33:34  kagel
   Changed #if defined(__AIX__) to #if defined( _AIX ).

   Revision 2.2  1999/10/04 15:59:47  kagel
   Fixed a few platform specific compilation problems affecting mainly
   AIX.  Check the #if defined( __AIX__ ) for correctness if this is your
   platform.

   Revision 2.1  1999/09/29 16:47:10  kagel
   Establishing current  source version id for initial working version of
   newly modularized source.  This is Feature Version 6.1 Source revision
   2.1.

   Revision 1.2  1999/09/22 20:21:41  kagel
   Moved statement preparation and cursor declaration in here.

   Revision 1.1  1999/09/01 19:46:53  kagel
   Initial revision


*/

static const char RCSHeader[] = "@(#)$Header: /home/art/GoogleDrive/Latest\040Art's\040Stuff/myschema.d/fork/RCS/print_indexes.ec,v 2.77 2023/12/11 14:21:26 art Exp $";
static const char RCSCompile[] = "@(#)$cc: " __FILE__ " " __DATE__ " " __TIME__ " $";

#if defined( _AIX ) /* Needed as AIX defines a locale control structure loc_t */
#if !defined( _H_LOCALEDEF )
#define _H_LOCALEDEF
#endif
#endif

#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <sys/types.h>
#include <string.h>
#include <limits.h>
#include <ctype.h>
#include <errno.h>

EXEC SQL INCLUDE sqlhdr;
EXEC SQL INCLUDE sqlca;
EXEC SQL INCLUDE sqltypes;
EXEC SQL INCLUDE datetime;
EXEC SQL INCLUDE locator;

EXEC SQL INCLUDE myschema.eh;
EXEC SQL INCLUDE prototypes.h;

static int prepped = 0;
static int FillFactor = 0;

/* MACRO */
#define limit_unit( a ) (a == 1 ? "KB" : (a == 1024 ? "MB" : "GB" ))
#define disposition( a ) (a == 8 ? "DISCARD" : "DETACH" )
#define which_disp( a ) ((((a >> 6) % 2) == 1) ? "INTERVAL ONLY" : ((((a >> 5) % 2) == 1) ? "ANY" : "INTERVAL FIRST"))

void print_indexstate( FILE *outfile, char *state );

void print_indexes( FILE *outfile, FILE *distfile, int SimpleFrags )
{
    int i, j;
    sortdir FuncDir, ColDir;
    char typetext[60], FieldName[129];
    char *p, *kp, *Key;
    keytype KeyType;
    char sortstrs[3][6] = {""," ASC"," DESC"};
    static char statement[8192], ER_Index[32767], Hldr[32767];
    int PrintParts = TRUE;
    int uknum = 1, fknum = 1, aknum = 1;;

    EXEC SQL BEGIN DECLARE SECTION;
    short am_ind, am_name_ind, opcl_ind, state_ind, keyind;
    int amid, defopclass;
    string am_name[IDBYTES], lastidx[IDBYTES], state[2], Partial[11];
EXEC SQL IFNDEF NOINT8;
    lvarchar amparam[4096];
    lvarchar keystring[4096];
EXEC SQL ELSE;
    varchar amparam[4096];
    varchar keystring[4096];
EXEC SQL ENDIF;
    string funcname[IDBYTES], opclassname[IDBYTES];
    EXEC SQL END DECLARE SECTION;

    /* Determine user's preferred FILLFACTOR. */
    if (!FillFactor) {
        p = getenv( "FILLFACTOR" );
        if (p) {
            FillFactor = atoi( p );
	}
    }
    
    /* Select for index information. */
    if (!prepped) {
        strcpy( statement,
                "SELECT i.idxname, i.owner, i.tabid, i.idxtype, "
                "i.clustered, " );
        if (Universal) {
            strcat( statement,
                    "\"informix\".indexkeyarray_out(i.indexkeys) AS allparts, " );
	    if (major_vers >= 12) {
		strcat( statement, "      nhashcols, nbuckets, fextsize, nextsize, indexattr, " );
	    } else if (major_vers == 11 && minor_vers >= 70) {
		strcat( statement, "      nhashcols, nbuckets, fextsize, nextsize, 0, " );
	    } else {
		strcat( statement, " 0, 0, 0, 0, 0, " );
	    }
        } else {
            strcat( statement,
                    "      part1, part2, part3, part4, part5, part6, " );
            strcat( statement,
                    "      part7, part8, part9, part10, part11, part12, " );
            strcat( statement, "      part13, part14, part15, part16, " );
	    strcat( statement, " 0, 0, 0, 0, 0, " );
        }
        strcat( statement, "      c.constrid, c.constrname, c.tabid, " );
        strcat( statement, "      c.constrtype, c.idxname, " );
        if (Universal) {
            strcat( statement, "i.amid, i.amparam, sa.am_name, " );
        } else {
            strcat( statement, "-1, '', '', " );
	}
        strcat( statement, "os.state " );
        if (Universal) {
            strcat( statement,
                    " FROM \"informix\".sysindices i, "
                    "       outer \"informix\".sysconstraints c, "
                    "       outer \"informix\".sysams sa, " );
        } else {
            strcat( statement,
                    "  FROM \"informix\".sysindexes i, "
                    "       outer \"informix\".sysconstraints c, " );
	}
        strcat( statement, " outer \"informix\".sysobjstate os " );
        strcat( statement, "  WHERE i.tabid = ? " );
        if (NoKeyIndices) {
            strcat( statement, "    AND i.idxname[1,1] != ' ' " );
	}
        strcat( statement,
                "    AND i.tabid = c.tabid AND i.idxname = c.idxname " );
        if (Universal) {
            strcat( statement,
                    "    AND i.amid = sa.am_id " );
	    if ((major_vers == 11 && minor_vers >= 50) || major_vers > 11) {
		strcat( statement,
			"    AND NOT EXISTS ("
			"        SELECT 1 "
			"        FROM syscolumns sc, sysindexes si "
			"        WHERE i.idxname = si.idxname "
			"          AND sc.tabid = si.tabid "
			"          AND si.part1 == sc.colno "
			"          AND sc.colattr == 17 "
			"          AND si.part4 = 0 ) " );
	    }
        }
        strcat( statement, "  AND os.tabid = i.tabid AND os.name = i.idxname AND os.objtype = 'I' " );
        strcat( statement, "  ORDER BY 1, c.constrtype, 3, 5 DESC;" );

        EXEC SQL PREPARE s_idx from :statement;
        if (sqlca.sqlcode) {
            fprintf( stderr,
                     "Cannot PREPARE select statement for indexes, "
                     "error: %d.\n",
                     sqlca.sqlcode );
            fprintf( stderr, "Error at character %d, value: %s.\nStatement is:\n",
		     sqlca.sqlerrd[4],
                     sqlca.sqlerrm );
            fprintf( stderr, "%s\n", statement );
            rexit(19);
        }
        EXEC SQL DECLARE idx CURSOR FOR s_idx;
        if (sqlca.sqlcode ) {
            fprintf( stderr,
                     "Cannot declare idx cursor; Error = %d\n", sqlca.sqlcode );
            rexit(20);
        }

        if (!conn5 && ids_server && !no_storage_opt) {
            char statement2[2048];

            snprintf( statement2,
                      sizeof statement2,
                      "SELECT fragtype, tabid, indexname, strategy, evalpos, "
		      "       nrows, npused, colno, flags, "
                      "       exprtext, dbspace, %s "
                      "FROM \"informix\".sysfragments "
                      "WHERE tabid = ? AND indexname = ? AND fragtype = 'I' "
                      "ORDER BY evalpos ",
                      major_vers >= 10 ? "partition" : "' '" );
            EXEC SQL PREPARE s_ifrags FROM :statement2;
            if (sqlca.sqlcode) {
                fprintf( stderr,
                         "Cannot PREPARE SELECT statement for index fragments, "
                         "error: %d.\n",
                         sqlca.sqlcode );
                fprintf( stderr, "Error at character %d.\nStatement is:\n",
                         sqlca.sqlerrd[4] );
                fprintf( stderr,
                         "%s\n",
                         statement2 );
                rexit(19);
            }
            EXEC SQL DECLARE ifrags CURSOR FOR s_ifrags;
            if (sqlca.sqlcode ) {
                fprintf( stderr,
                         "Cannot DECLARE ifrags CURSOR; Error = %d\n",
                         sqlca.sqlcode );
                rexit(20);
            }

            ifragments.exprtext.loc_loctype = LOCUSER;
            ifragments.exprtext.loc_buffer = (char *)malloc( 4096 );
            if (!ifragments.exprtext.loc_buffer) {
                fprintf( stderr,
                         "Cannot allocate memory for index fragment "
                         "expressions.  Errno= %d.\n",
                         errno );
                rexit(99);
            }
            ifragments.exprtext.loc_bufsize = 4096;
            ifragments.exprtext.loc_write = writeit;
            ifragments.exprtext.loc_open = openit;
            ifragments.exprtext.loc_close = closeit;
            iblobbuff.bbsize = 2048;
            iblobbuff.buff = (char *)malloc( iblobbuff.bbsize );
            if (!iblobbuff.buff) {
                fprintf( stderr,
                         "Cannot allocate memory for index data "
                         "expressions.  Errno= %d.\n",
                         errno );
                rexit(99);
            }
            iblobbuff.global_length = 0;
        }
        if (Universal) {
            EXEC SQL PREPARE s_ixfuncname FROM
                "SELECT procname FROM \"informix\".sysprocedures where procid = ?";
            if (sqlca.sqlcode) {
                fprintf( stderr,
                         "Cannot PREPARE select statement for index funcname, "
                         "error: %d.\n",
                         sqlca.sqlcode );
                rexit(19);
            }
            EXEC SQL PREPARE s_funcopclass FROM
                "SELECT oc.opclassname, sa.am_defopclass \
                 FROM \"informix\".sysopclasses oc, outer \"informix\".sysams sa \
                 WHERE oc.amid = sa. am_id \
                   AND oc.opclassid = ?";
            if (sqlca.sqlcode) {
                fprintf( stderr,
                         "Cannot PREPARE select statement for opclass, "
                         "error: %d.\n",
                         sqlca.sqlcode );
                rexit(19);
            }
        }
        prepped = 1;
    }

    lastidx[0] = (char)0;
    EXEC SQL OPEN idx USING :prev.tabid;
    if (sqlca.sqlcode) {
        fprintf( stderr,
                 "Error opening cursor idx; Error = %d\n",
                 sqlca.sqlcode );
        rexit(14);
    }
    while (!sqlca.sqlcode) {
	memset( &sysindexes, 0, sizeof sysindexes );
        if (!Universal) {
            EXEC SQL
                FETCH idx INTO
                :sysindexes.idxname,
                :sysindexes.owner,
                :sysindexes.tabid,
                :sysindexes.idxtype,
                :sysindexes.clustered,
                :sysindexes.part[0],
                :sysindexes.part[1],
                :sysindexes.part[2],
                :sysindexes.part[3],
                :sysindexes.part[4],
                :sysindexes.part[5],
                :sysindexes.part[6],
                :sysindexes.part[7],
                :sysindexes.part[8],
                :sysindexes.part[9],
                :sysindexes.part[10],
                :sysindexes.part[11],
                :sysindexes.part[12],
                :sysindexes.part[13],
                :sysindexes.part[14],
                :sysindexes.part[15],
		:sysindexes.nhashcols,
		:sysindexes.nbuckets,
		:sysindexes.fextsize,
		:sysindexes.nextsize,
		:sysindexes.indexattr,
                :sysconstraints.constrid INDICATOR isconstraint,
                :sysconstraints.constrname,
                :sysconstraints.tabid,
                :sysconstraints.constrtype,
                :sysconstraints.idxname,
                :sysindexes.amid,
                :amparam INDICATOR am_ind,
                :am_name INDICATOR am_name_ind,
                :state;
        } else {
            EXEC SQL
                FETCH idx INTO
                :sysindexes.idxname,
                :sysindexes.owner,
                :sysindexes.tabid,
                :sysindexes.idxtype,
                :sysindexes.clustered,
                :keystring INDICATOR keyind,
		:sysindexes.nhashcols,
		:sysindexes.nbuckets,
		:sysindexes.fextsize,
		:sysindexes.nextsize,
		:sysindexes.indexattr,
                :sysconstraints.constrid INDICATOR isconstraint,
                :sysconstraints.constrname,
                :sysconstraints.tabid,
                :sysconstraints.constrtype,
                :sysconstraints.idxname,
                :sysindexes.amid,
                :amparam INDICATOR am_ind,
                :am_name INDICATOR am_name_ind,
                :state INDICATOR state_ind;
        }
	/* I have found orphaned indexes.  Disabled indexes with a generated
	   name that were created by a constraint, but the constraint was no
	   longer in existence.  Skip them here. */
	if (state_ind == 0 && state[0] == 'D' && sysindexes.idxname[0] == ' ' && isconstraint == -1) {
	    continue;
	}
	if (replace_owners) {
	  strcpy( sysindexes.owner, owner_override );
	}
        if (sqlca.sqlcode) {
            break;
        }
        if (strcmp( sysindexes.idxname, lastidx ) != 0) {
            strcpy( lastidx, sysindexes.idxname );
            if (isconstraint == 0 && sysindexes.idxname[0] == ' ') {
		/* New -K option.  Name unnamed indexes after table & type. */
		if (LongKeyIndices) {
		    char ext[20];

		    if (sysconstraints.constrtype[0] == 'R') {
			snprintf( ext, sizeof ext, "fk%.3d", fknum++ );
		    } else if (sysconstraints.constrtype[0] == 'U') {
			snprintf( ext, sizeof ext, "uk%.3d", uknum++ );
		    } else if (sysconstraints.constrtype[0] == 'P') {
			snprintf( ext, sizeof ext, "pk" );
		    } else {
			snprintf( ext, sizeof ext, "ak%.3d", aknum++ );
		    }
		    snprintf( sysindexes.idxname,
			      sizeof sysindexes.idxname,
			      "%.107s_%s",
			      prev.tabname,
			      ext );
		} else {
		    for (p = sysindexes.idxname; *p == ' '; p++)
			*p = sysconstraints.constrtype[0];
		    /* Name should be stripped and null terminated, but to be sure. */
		    strtok( sysindexes.idxname, " " );
		    fprintf( outfile,
			     "-- Index <%s> is a constraint index.  "
			     "Creating normal index: %s.\n",
			     lastidx,
			     sysindexes.idxname );
		}
            } else {
                strtok( sysindexes.idxname, " " );
	    }
            typetext[0] = ' ';
            typetext[1] = (char)0;
            if (sysindexes.idxtype[0] == 'U') {
                strcat( typetext, "UNIQUE " );
	    }
            if (sysindexes.clustered[0] == 'C') {
                strcat( typetext, "CLUSTER " );
	    }
            if (!stripowner) {
                fprintf( outfile,
                         "CREATE%sINDEX%s \"%s\".%s ON \"%s\".%s (\n\t",
                         typetext,
			 create_conditionally ? " IF NOT EXISTS" : "",
                         sysindexes.owner,
                         sysindexes.idxname,
                         prev.owner,
                         prev.tabname );
		if (sysconstraints.constrtype[0] == 'P' && do_For_ER == 1) {
		    snprintf( ER_Index,
			      sizeof ER_Index,
			      "CREATE%sINDEX%s \"%s\".%s_for_er ON \"%s\".%s (\n\t",
			      typetext,
			      create_conditionally ? " IF NOT EXISTS" : "",
			      sysindexes.owner,
			      sysindexes.idxname,
			      prev.owner,
			      prev.tabname );
		}
            } else {
                fprintf( outfile,
                         "CREATE%sINDEX%s %s ON %s (\n\t",
                         typetext,
			 create_conditionally ? " IF NOT EXISTS" : "",
                         sysindexes.idxname,
                         prev.tabname );
		if (sysconstraints.constrtype[0] == 'P' && do_For_ER == 1) {
		    snprintf( ER_Index,
			      sizeof ER_Index,
			      "CREATE%sINDEX%s %s_for_er ON %s (\n\t",
			      typetext,
			      create_conditionally ? " IF NOT EXISTS" : "",
			      sysindexes.idxname,
			      prev.tabname );
		}
	    }
            kp = keystring;
            if (Universal) {
                if (keyind > 0) {
                    fprintf( stderr,
                             "Key information too long (%d)!  "
                             "Some keys may not be processed.\n",
                             keyind );
		}
            }
            FuncDir = Asc;
            if (Universal) {
                kp = keystring;
            } else {
                KeyType = ColKey;
	    }
            for (i=0; i<16; i++) {
                /* Up to 16 keys of any type! */
                char opclsstr[IDBYTES+10];
                int first, col;

                if (Universal) {
                    kp = NextKey( kp, &KeyType, &Key );
                    if (KeyType == NoKey) {
                        break;  /* No more keys. */
                    } else if (KeyType == ColKey) {
                        KeyCol( Key,
                                &sysindexes.part[i],
                                &sysindexes.op_class_id[i],
                                &ColDir );
                        j = abs( sysindexes.part[i] ) - 1;
                    } else if (KeyType == FuncKey) {
                        int procid, col;

                        sysindexes.funcid = FuncId( Key );
			if (sysindexes.funcid > 0) {
			    EXEC SQL EXECUTE s_ixfuncname
				INTO :funcname
				USING :sysindexes.funcid;
			    if (sqlca.sqlcode) {
				fprintf( stderr,
					 "Error EXECUTING index function name "
					 "query, error: %d, procid=%d.\n",
					 sqlca.sqlcode,
					 sysindexes.funcid);
				rexit(19);
			    }
			} else if (sysindexes.funcid < 0) {
			    switch (sysindexes.funcid) {
			    case -1660:
				strcpy( funcname, "bson_value_double" );
				break;
			    case -1661:
				strcpy( funcname, "bson_value_lvarchar" );
				break;
			    case -1666:
				strcpy( funcname, "bson_value_boolean" );
				break;
			    case -1667:
				strcpy( funcname, "bson_value_date" );
				break;
			    case -1669:
				strcpy( funcname, "bson_value_int" );
				break;
			    case -1670:
				strcpy( funcname, "bson_value_bigint" );
				break;
			    case -1671:
				strcpy( funcname, "bson_value_timestamp" );
				break;
			    case -1678:
				strcpy( funcname, "bson_value_varchar" );
				break;
			    default:
				snprintf( funcname, sizeof funcname, "Unknown bson funcid: %d.\n", sysindexes.funcid );
				break;
			    }
			}

                        fprintf( outfile, "%s%s( ",
                                 i>0 ? ", \n\t" : "",
                                 funcname );
                        FuncDir = Asc;
                        first = 1;

			do {
			    memset( FieldName, 0, sizeof FieldName );
			    col = FuncCol( &ColDir, FieldName );
                            if (ColDir == Desc) {
                                FuncDir = Desc;
			    }
                            if (!first && (col || strlen(FieldName))) {
                                fprintf( outfile, ",\n\t\t" );
                            } else {
                                first = 0;
                            }
			    if (col) {
				fprintf( outfile,
					 " %s",
					 columns[abs( col ) - 1] );
			    } else if (strlen( FieldName )) {
				fprintf( outfile, " %s", FieldName );
			    }
                        } while (col);

                        fprintf( outfile, " )" );
                        FuncOpCl( &sysindexes.op_class_id[i] );
                        EXEC SQL EXECUTE s_funcopclass 
                            INTO :opclassname, :defopclass INDICATOR opcl_ind
                            USING :sysindexes.op_class_id[i];
                        if (sqlca.sqlcode) {
                            fprintf( stderr,
                                     "Error EXECUTING opclas name query, "
                                     "error: %d, class_id=%d.\n",
                                     sqlca.sqlcode,
                                     sysindexes.op_class_id[i] );
                            rexit(19);
                        }
                        if (opcl_ind < 0) {
                            defopclass = -1;/* Another Excalibur Kludge. */
			}
                        if (defopclass == sysindexes.op_class_id[i]) {
                            opclsstr[0] = (char)0;
                        } else {
                            snprintf( opclsstr,
				      sizeof opclsstr,
                                     " %s ",
                                     opclassname );
			}
                        if (Universal && sysindexes.amid != 1) {
                            FuncDir = NoDir;
			}
                        if (Universal && sysindexes.funcid) {
                            fprintf( outfile,
                                     "%s",
                                     sortstrs[FuncDir] );
                        }
                        continue;
                    }
                } else {
                    KeyType = ColKey;
                    j = abs(sysindexes.part[i]) - 1;
                    ColDir = sysindexes.part[i] < 0 ? Desc : Asc;
                }
                if (KeyType == ColKey) {
                    if (sysindexes.part[i] == 0) {
                        break;  /* No more keys. */
		    }
                    if (i) {
                        fprintf( outfile, ",\n\t" );
		    }
                    if (Universal && sysindexes.op_class_id[i] > 0) {
                        EXEC SQL EXECUTE s_funcopclass 
                            INTO :opclassname, :defopclass INDICATOR opcl_ind
                            USING :sysindexes.op_class_id[i];
                        if (sqlca.sqlcode) {
                            fprintf( stderr,
                                     "Cannot EXECUTING opclas name query, "
                                     "error: %d.\n",
                                     sqlca.sqlcode );
                            rexit(19);
                        }
                        if (opcl_ind < 0) {
                            defopclass = -1;/* Another Excalibur Kludge. */
			}
                        if (defopclass == sysindexes.op_class_id[i]) {
                            opclsstr[0] = (char)0;
                        } else {
                            snprintf( opclsstr,
				      sizeof opclsstr,
                                     " %s",
                                     opclassname );
			}
                    } else {
                        opclsstr[0] = (char)0;
		    }
                    if (Universal && sysindexes.amid != 1)
                        ColDir = NoDir;
                    snprintf( Hldr, sizeof Hldr,
                             "%s%s%s",
                             columns[j],
                             opclsstr,
                             sortstrs[ColDir] );
                    fprintf( outfile, "%s", Hldr );
		    if (sysconstraints.constrtype[0] == 'P' && do_For_ER == 1) {
			strcat( ER_Index, Hldr );
		    }
                } else if (Universal) {
                    if (ColDir == Desc) {
                        FuncDir = Desc;
		    }
                    if (i) {
                        fprintf( outfile, ",\n\t\t" );
		    }
		    snprintf( Hldr, sizeof Hldr, "%s", columns[j] );
                    fprintf( outfile, "%s", Hldr );
		    if (sysconstraints.constrtype[0] == 'P' && do_For_ER == 1) {
			strcat( ER_Index, Hldr );
		    }
                }
            }
            fprintf( outfile, "\n)" );
	    if (sysconstraints.constrtype[0] == 'P' && do_For_ER == 1)
		strcat( ER_Index, ",\n\tifx_replcheck\n)" );
            if (sysindexes.amid > 0 ) {
                /* KLUDGE!  Seems Excalibur DataBlade does not insert a sysams
                   record for this method!  Somehow dbschema understands!??! */
                if (am_name_ind < 0) {
                    if (sysindexes.amid == 100) {
                        strcpy( am_name, "etx" );
                    } else {
                        snprintf( am_name,
                                  IDBYTES,
                                  "{UnRegistered Access Method, am_id: %d}",
                                  sysindexes.amid );
		    }
		}
                fprintf( outfile, " USING %s", am_name );
                if (am_ind >= 0) {
                    fprintf( outfile, " (%s)", amparam );
                    if (am_ind > 0) {
                        fprintf( outfile,
                                 " { Access Method parameters have been "
                                 "truncated! }" );
		    }
                }
            }
            if (FillFactor) {
                snprintf( Hldr, sizeof Hldr, " FILLFACTOR %d", FillFactor );
                fprintf( outfile, "%s", Hldr );
		if (sysconstraints.constrtype[0] == 'P' && do_For_ER == 1) {
		    strcat( ER_Index, Hldr );
		}
	    }
            if (!conn5 && ids_server) {
                blobbuff = &iblobbuff;
		if (!no_storage_opt) {
		    int nfrags = 0;

		    EXEC SQL OPEN ifrags USING :prev.tabid, :lastidx;
		    if (sqlca.sqlcode < 0) {
			fprintf( stderr,
				 "%d: Cannot open CURSOR ifrags.  Error = %d\n",
				 __LINE__,
				 sqlca.sqlcode );
			rexit(1);
		    }
		    while (sqlca.sqlcode == 0) {
			int nrows = 0, npused = 0, colno = 0, flags = 0;

			EXEC SQL FETCH ifrags INTO :ifragments;
			if (sqlca.sqlcode < 0) { // Mark
			    fprintf( stderr,
				     "Cannot FETCH CURSOR ifrags.  Error = %d\n",
				     sqlca.sqlcode );
			    rexit(1);
			} else if (sqlca.sqlcode == SQLNOTFOUND) {
			    if (nfrags == 0) {
				/* Index is a legacy attached index or explicitely created semi-attached! */
				fprintf( outfile, " IN TABLE" );
			    }
			    break;
			}
			nfrags++;

			if (SimpleFrags && strcmp( ifragments.dbspace, ifragments.partition ) == 0) {
			    PrintParts = FALSE;
			} else {
			    PrintParts = TRUE;
			}
			if (ifragments.flags & 0x8000000) {
			    strcpy( Partial, " INDEX OFF" );
			} else {
			    Partial[0] = (char)0;
			}
			if (ifragments.strategy[0] == 'I') {
			    /* Index in a single dbspace. */
			    snprintf( Hldr, sizeof Hldr,
				     " IN %s",
				     ifragments.dbspace );
			    fprintf( outfile, "%s", Hldr );
			    if (sysconstraints.constrtype[0] == 'P' && do_For_ER == 1) {
				strcat( ER_Index, Hldr );
			    }
			} else if (ifragments.strategy[0] == 'L') {
			    /* List Fragmentation, initial list column definition. */
			    if (ifragments.evalpos < 0) {
				snprintf( Hldr, sizeof Hldr,
					 " FRAGMENT BY LIST (%s) \n",
					 iblobbuff.buff );
				fprintf( outfile, "%s", Hldr );
				if (sysconstraints.constrtype[0] == 'P' && do_For_ER == 1) {
				    strcat( ER_Index, Hldr );
				}
			    } else {
				/* List fragmentation, partition list. */
				if (ifragments.evalpos > 0) {
				    fprintf( outfile, ", \n" );
				}
				snprintf( Hldr, sizeof Hldr,
					 "  PARTITION %s %s IN %s%s",
					 ifragments.partition,
					 iblobbuff.buff,
					 ifragments.dbspace,
					 Partial );
				fprintf( outfile, "%s", Hldr );
				if (sysconstraints.constrtype[0] == 'P' && do_For_ER == 1) {
				    strcat( ER_Index, Hldr );
				}
			    }
			    continue;
			} else if (ifragments.strategy[0] == 'N') {
			    /* Range/Interval fragmentation, Interval Definition. */
			    if (ifragments.evalpos < 0) {
				switch (ifragments.evalpos) {
				case -4:
				    /* Vers 12.10+ Rolling Partitions. */
				    nrows = ifragments.nrows;
				    npused = ifragments.npused;
				    flags = ifragments.flags;
				    colno = ifragments.colno;
				    break;
				case -3:
				    fprintf( outfile,
					     " FRAGMENT BY RANGE( %s )\n",
					     iblobbuff.buff );
				    break;
				case -2:
				    fprintf( outfile,
					     "  INTERVAL( %s )",
					     iblobbuff.buff );
				    if (colno > 0) {
					fprintf( outfile,
						 "\n  ROLLING ( %d FRAGMENTS )",
						 colno );
				    }
				    if (flags == 8) {
					fprintf( outfile, " DISCARD" );
				    } else if (flags == 16) {
					fprintf( outfile, " DETACH" );
				    }
				    if (nrows > 0) {
					fprintf( outfile,
						 " LIMIT TO %d %s %s %s\n",
						 nrows,
						 limit_unit( npused ),
						 disposition( flags ),
						 which_disp( flags ) );
				    }
				    break;
				case -1:
				    if (((ifragments.flags >> 15) % 2)  == 1) {
					fprintf( outfile,
						 "\n  STORE IN ( %s() )\n",
						 iblobbuff.buff );
				    } else {
					fprintf( outfile,
						 "\n  STORE IN ( %s )\n",
						 iblobbuff.buff );
				    }
				    break;
				}
			    } else {
				if (ifragments.evalpos > 0) {
				    fprintf( outfile, ", \n" );
				}
				snprintf( Hldr,
					  sizeof Hldr,
					 "%s  PARTITION %s %s IN %s%s",
					 (ifragments.evalpos > 0) ? ", \n" : "",
					 ifragments.partition,
					 iblobbuff.buff,
					 ifragments.dbspace,
					 Partial );
				fprintf( outfile, "%s", Hldr );
				if (sysconstraints.constrtype[0] == 'P' && do_For_ER == 1) {
				    strcat( ER_Index, Hldr );
				}
			    }
			    continue;
			} else if (ifragments.strategy[0] == 'R') {
			    /* Indexes cannot be fragmented by ROUND ROBIN, but
			       keep this for consistency. */
			    if (ifragments.evalpos == 0) {
				if (major_vers >= 10 && PrintParts) {
				    fprintf( outfile,
					     " FRAGMENT BY ROUND ROBIN \n\t" );
				} else {
				    fprintf( outfile,
					     " FRAGMENT BY ROUND ROBIN IN \n\t" );
				}
				if (major_vers >= 10 && PrintParts) {
				    fprintf( outfile,
					     "PARTITION %s IN %s",
					     ifragments.partition,
					     ifragments.dbspace );
				} else {
				    fprintf( outfile,
					     "%s",
					     ifragments.dbspace );
				}
			    } else {
				fprintf( outfile, ", \n\t" );
				if (major_vers >= 10 && PrintParts) {
				    fprintf( outfile,
					     "PARTITION %s IN %s",
					     ifragments.partition,
					     ifragments.dbspace );
				} else {
				    fprintf( outfile,
					     "%s",
					     ifragments.dbspace );
				}
			    }
			} else if (ifragments.strategy[0] == 'E') {
			    /* Expression fragmentation. */
			    if (ifragments.evalpos == 0) {
				if (major_vers >= 10 && PrintParts) {
				    snprintf( Hldr, sizeof Hldr,
					     " FRAGMENT BY EXPRESSION \n   PARTITION %s %s IN %s%s",
					     ifragments.partition,
					     iblobbuff.buff,
					     ifragments.dbspace,
					     Partial );
				} else {
				    snprintf( Hldr, sizeof Hldr,
					     " FRAGMENT BY EXPRESSION \n  %s IN %s%s",
					     iblobbuff.buff,
					     ifragments.dbspace,
					     Partial );
				}
				fprintf( outfile, "%s", Hldr );
			    } else {
				if (major_vers >= 10 && PrintParts) {
				    snprintf( Hldr, sizeof Hldr,
					     ", \n   PARTITION %s %s IN %s%s",
					     ifragments.partition,
					     iblobbuff.buff,
					     ifragments.dbspace,
					     Partial );
				} else {
				    snprintf( Hldr, sizeof Hldr,
					      ", \n  %s IN %s%s",
					      iblobbuff.buff,
					      ifragments.dbspace,
					      Partial );
				}
				fprintf( outfile, "%s", Hldr );
			    }
			    if (sysconstraints.constrtype[0] == 'P' && do_For_ER == 1) {
				strcat( ER_Index, Hldr );
			    }
			}
		    }
		    EXEC SQL CLOSE ifrags;
		}

		/* Print out any FOT Hash clause. */
		if (sysindexes.nhashcols > 0) {
		    fprintf( outfile, "\nHASH ON ( \n" );
		    for (j=0; j<sysindexes.nhashcols; j++) {
			if (j > 0)
			    fprintf( outfile, ", \n" );
			fprintf( outfile,
				 "\t%s",
				 columns[j] );
		    }
		    fprintf( outfile,
			     "\n) WITH %d BUCKETS",
			     sysindexes.nbuckets );
		}

		/* Print out any non-default extent settings. */
		if (sysindexes.fextsize > 0) {
		    fprintf( outfile,
			     "\nEXTENT SIZE %d NEXT SIZE %d",
			     sysindexes.fextsize,
			     sysindexes.nextsize );
		}
            }
	    if ((isconstraint == -1 || state[0] != 'D') && state_ind == 0)
		print_indexstate( outfile, state );
	    if (do_index_online && (sysindexes.indexattr & 4) == 0)
		fprintf( outfile, " ONLINE" );

	    /* Print out if index is created compressed. */
	    if ((sysindexes.indexattr & 2))
		fprintf( outfile, " COMPRESSED" );

	    /* Print out if the index is a BSON collection index. */
	    if ((sysindexes.indexattr & 4))
		fprintf( outfile, " COLLECTION ''" );

	    fprintf( outfile, ";\n\n" );
	    /* Print out the ER Index if requested. */
	    if (sysconstraints.constrtype[0] == 'P' && do_For_ER == 1) {
		fprintf( outfile, "%s;\n\n", ER_Index );
		Hldr[0] = (char)0; /* clear the ER Index buffer. */
	    }
            if (doDistrib) { 
                fprintf( distfile,
                         "UPDATE STATISTICS%s FOR TABLE %s ( \n",
                         conn5 ? "" : " LOW",
                         prev.tabname );
                for (i=0; i<16 && sysindexes.part[i]; i++) {
                    j = abs(sysindexes.part[i]) - 1;
                    if (i)
                        fprintf( distfile, ",\n" );
                    fprintf( distfile,
                             "\t%s",
                             columns[j] );
                }
                fprintf( distfile, " );\n" );
            }
            fprintf( outfile, "\n" );
        }
    }
    if (sqlca.sqlcode && sqlca.sqlcode != SQLNOTFOUND) {
        fprintf( stderr,
                 "Error fetching index information. Error = %d\n",
                 sqlca.sqlcode );
        fprintf( stderr,
                 "\tsqlerrp = %s\n\tsqlerrm = %s\n",
                 sqlca.sqlerrp,
                 sqlca.sqlerrm );
        rexit(14);
    }
    
    EXEC SQL CLOSE idx;
}

/*
 * IDS 9.3 docs say the indexkey contains a string of the form:
 * 'funcid (col1, col2, col3,...) [op class]'  
 * Given that an index can have more than one functional key???, and that each
 * functional key can pass multiple columns and also that any and all
 * keys can specify an op clasee, shouldn't that be:
 * 'funcid1 (arg1_1,arg1_2,...)[op cl1], funcid2 (arg2_1, arg2_2,...)[op cl2]'
 * ??
 * 8/12/03: Actually its:
 * 'keyspec, keyspec, keyspec,....'
 * where 'keyspec' is either a funcspec or a colspec where funcspec is:
 * '<procid>(col1no, col2no, col3no,...) [opclsid]' and colspec is:
 * 'colno [op_class]'
 * Note if the function key is defined as DESC in the index definition the first 
 * colno in the funcspec will be negative.  Column and functional keys may be freely 
 * mixed in an index definition.
 *
 * If a normal (column key only) index the format is:
 * colno [op_class], ....
 * This is repeated for each key.
 */

static char *tok = (char *)NULL, *end;
static int done = 0;

typedef enum { Init, Scan, Bracket, Comma, End } scanstate;

char *NextKey( char *str, keytype *type, char **key )
{
    char *p = str, *next;
    scanstate state = Init;

    if (*p == (char)0) {
        *type = NoKey;
        return p;
    } else if (*p == '<')
        *type = FuncKey;
    else
        *type = ColKey;

    while (state != End) {
        switch (state) {
        case Init:
            if (*p == (char)0) {
                *type = NoKey;
                next = p;
                state = End;
                break;
            } else 
                state = Scan; /* Fall through OK. */
        case Scan:
            if (*p == ']')
                state = Bracket;
            break;
        case Bracket:
            if (*p == ',' )
                state = Comma;
            else if (*p == (char)0) {
                state = End;
                next = p;
            }
            break;
        case Comma:
            if (*p == ' ') {
                state = End;
                next = p + 1;
            } else if (*p == (char)0) {
                state = End;
                next = p;
            }
        }
        p++;
    }
    *key = str;
    tok = (char *)NULL;
    done = 0;

    return next;
}

void KeyCol( char *key, short *col, int *opcl, sortdir *ColDir )
{
    char *p = key;

    if (done) {
        *col = 0;
        return;
    }
    if (!tok) {
        tok = key;
        while (*tok && (*tok == ' ' || *tok == ','))
            tok++;
    }
    *col = atoi( tok );
    while (*tok && *tok != '[')
        tok++;
    tok++;
    *opcl = atoi( tok );
    while (*tok && *tok != ']')
        tok++;
    tok++;
    if (*tok == (char)0)
        done = 1;
    tok++;
    if (*col < 0)
        *ColDir = Desc;
    else
        *ColDir = Asc;

    return;
}

int FuncId( char *key )
{
    char *p = key;
    int id;

    while (*p && *p != '<')
        p++;
    p++;
    id = atoi( p );
    while (*p && *p != '>')
        p++;
    tok = p;

    return id;
}

int FuncCol( sortdir *ColDir, char *FieldName )
{
    int col;

    if (done) {
        return 0;
    }
    while (*tok && *tok != '(' && *tok != ',')
        tok++;
    tok++;
    col = atoi( tok );

    if (col) {
	if (col < 0)
	    *ColDir = Desc;
	else
	    *ColDir = Asc;
    } else {
	char *p = tok, *P = FieldName;
	int del;

	/* Named bson field or constant instead of numbered column. */
	while (*p == ' ')
	    p++;
	if (*p == '\'')
	    del = 1;
	else
	    del = 0;
	do {
	    *P = *p;
	    p++; P++;
	} while (!((del && *p == '\'') || (!del && (*p == ',' || *p == ')'))));
	if (del && *p == '\'')
	    *P = *p;
	*(++P) = (char)0;
    }
    while (*tok && *tok != ',' && *tok != ')')
	tok++;
    if (!(*tok) || *tok == ')')
	done = 1;

    return col;
}

void FuncOpCl( int *opcl )
{
    char *p = tok;
    while (*p && *p != '[')
        p++;
    p++;
    if (*p == (char)0) {
        *opcl = 0;
        return;
    }
    *opcl = (int)atoi( p );

    return;
}


void print_indexstate( FILE *outfile, char *state )
{
    switch (state[0]) {
      case 'D':
	  fprintf( outfile, " DISABLED" );
	  break;
      case 'F':
	  fprintf( outfile, " FILTERING WITHOUT ERRORS" );
	  break;
      case 'G':
	  fprintf( outfile, " FILTERING WITH ERRORS" );
	  break;
      case 'E':
	  break;
    }
}
