/* print_puts.ec - Print the PUT clause in CREATE TABLE statements to place 
   Smart BLOB columns in sdbdpaces.
*/
/*
   RCS Header:
   -----------
   $Header: /home/art/GoogleDrive/Latest\040Art's\040Stuff/myschema.d/fork/RCS/print_puts.ec,v 1.14 2023/12/13 23:10:59 art Exp $

   RCS Log:
   --------
   $Log: print_puts.ec,v $
   Revision 1.14  2023/12/13 23:10:59  art
   Fixed LOB locator option printout.

   Revision 1.13  2023/02/21 14:44:09  art
   Fixed the fix

   Revision 1.12  2023/02/19 17:08:47  art
   Fixed issue correctly reporting PUT clauses for BLOB and CLOB
   type columns in tables with CRCOLS defined.

   Revision 1.11  2021/02/17 17:15:33  art
   Fixed PUT clauses for smartblob types.

   Revision 1.10  2010/08/25 13:52:14  art
   Fixed several compile time warnings.  Thanks to John Adamski for identifying them.

   Revision 1.9  2010/01/05 23:43:11  art
   Fixed extra spaces after the column name in the PUT clause.

   Revision 1.8  2009/12/10 17:53:56  art
   Fixed problem with garbage in CREATE PROCEDURE lines.

   Revision 1.7  2009/02/11 18:30:34  art
   Fixed problem with LVARCHARS on AIX-64bit.

   Revision 1.6  2009/01/01 17:58:25  art
   Removed RCSWhat ident string which was BB specific.

   Revision 1.5  2004/10/18 19:32:16  kagel
     -- Fixed bug for Smart Blobs specified without EXTENT SIZE option
   Added support for remaining Smart Blob options (I think, docs are
   onconsistent.  Ex: says for one option flag == 0x00010000 == 16!  I assume
   the decimal representation is accurate.

   Revision 1.4  2004/09/14 17:17:51  kagel
   Fixing ANSI database access.

   Revision 1.3  2002/07/26 16:09:50  kagel
   Fixed sbspace query typo.

   Revision 1.2  2001/11/09 16:12:53  kagel
   Replacing dummy function with functional one.

   Revision 1.1  2001/11/08 21:58:00  kagel
   Initial revision


*/

static const char RCSHeader[] = "@(#)$Header: /home/art/GoogleDrive/Latest\040Art's\040Stuff/myschema.d/fork/RCS/print_puts.ec,v 1.14 2023/12/13 23:10:59 art Exp $";
static const char RCSCompile[] = "@(#)$cc: " __FILE__ " " __DATE__ " " __TIME__ " $";

#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>

EXEC SQL INCLUDE "myschema.eh";

EXEC SQL INCLUDE prototypes.h;


/*
 * Return 1 if clause was written, 0 otherwise so caller can format
 * accordingly. 
 */
int print_puts( FILE *outfile, int CrColFlag )
{
    static long prepped = 0;
    EXEC SQL BEGIN DECLARE SECTION;
    string colname[IDBYTES];
    int extentsize, flags;
    string sbspace[IDBYTES];
    int add_on = 0;
    EXEC SQL END DECLARE SECTION;
    int printed = 0, extras = 0;

    if ((prev.flags & CrColFlag))
	add_on = 2;

    if (!prepped) {

	EXEC SQL PREPARE s_slobs FROM
	    "SELECT c.colname, a.extentsize, a.flags, a.sbspace \
		FROM \"informix\".syscolattribs a, \"informix\".syscolumns c \
		WHERE a.tabid = c.tabid
		  AND a.colno = c.colno + ?
		  AND a.tabid = ?";
	if (sqlca.sqlcode < 0) {
	    fprintf( stderr,
		     "Error PREPARING sbspace query. Code=%d ISAM=%d.\n",
		     sqlca.sqlcode,
		     sqlca.sqlerrd[1] );
	    exit( 1 );
	}
	EXEC SQL DECLARE c_slobs CURSOR FOR s_slobs;
	if (sqlca.sqlcode < 0) {
	    fprintf( stderr,
		     "Error DECLARING sbspace cursor. Code=%d ISAM=%d.\n",
		     sqlca.sqlcode,
		     sqlca.sqlerrd[1] );
	    exit( 1 );
	}
	prepped = 1;
    }
    EXEC SQL OPEN c_slobs USING :add_on, :prev.tabid;
    if (sqlca.sqlcode < 0) {
	fprintf( stderr,
		 "Error OPENING sbspace query. Code=%d ISAM=%d.\n",
		 sqlca.sqlcode,
		 sqlca.sqlerrd[1] );
	exit( 1 );
    }
    
    for (;;) {
	extras = 0;
	EXEC SQL FETCH c_slobs INTO :colname, :extentsize, :flags, :sbspace;
	if (sqlca.sqlcode == SQLNOTFOUND)
	    break;
	else if (sqlca.sqlcode < 0) {
	    fprintf( stderr,
		     "Error FETCHING sbspace query. Code=%d ISAM=%d.\n",
		     sqlca.sqlcode,
		     sqlca.sqlerrd[1] );
	    exit( 1 );
	}
	if (!printed)
	    fprintf( outfile, " PUT" );
	fprintf( outfile,
		 "%s %s IN ( %s )",
		 printed ? "," : "",
		 colname,
		 sbspace );
	if (extentsize > 0) {
	    /* If EXTENT SIZE clause not specified at creation/alter time,
	       extentsize appears to be -1.  Guess I can live with doing
	       that also. */
	    fprintf( outfile,
		     "%sEXTENT SIZE %d",
		     !extras ? " (" : "",
		     extentsize );
	    extras++;
	}
	if (flags & LO_CREATE_LOG) {
	    fprintf( outfile,
		     "%sLOG",
		     !extras ? " (" : ", " );
	    extras++;
	}
	if (flags & LO_CREATE_NOLOG) {
	    fprintf( outfile,
		     "%sNO LOG",
		     !extras ? " (" : ", " );
	    extras++;
	}
	if (flags & LO_CREATE_KEEP_LASTACCESS_TIME) {
	    fprintf( outfile,
		     "%sKEEP ACCESS TIME",
		     !extras ? " (" : ", " );
	    extras++;
	}
	if (flags & LO_CREATE_NOKEEP_LASTACCESS_TIME) {
	    fprintf( outfile,
		     "%sNO KEEP ACCESS TIME",
		     !extras ? " (" : ", " );
	    extras++;
	}
	if (flags & LO_CREATE_HIGH_INTEG) {
	    fprintf( outfile,
		     "%sHI_INTEG",
		     !extras ? " (" : ", " );
	    extras++;
	}
	if (flags & LO_CREATE_MODERATE_INTEG) {
	    fprintf( outfile,
		     "%sMODERATE INTEG",
		     !extras ? " (" : ", " );
	    extras++;
	}
	if (extras)
	    fprintf( outfile, ") " );
	printed = 1;
    }
    EXEC SQL CLOSE c_slobs;

    return printed;
}
